"""
Divide and Conquer
------------------
"""

import gc
import logging
import os
from dataclasses import dataclass
from dataclasses import replace
from math import ceil
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import psutil  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut

from .collect import collect_metacells
from .direct import compute_direct_metacells

__all__ = [
    "set_max_parallel_piles",
    "get_max_parallel_piles",
    "guess_max_parallel_piles",
    "compute_target_pile_size",
    "divide_and_conquer_pipeline",
    "compute_divide_and_conquer_metacells",
]


MAX_PARALLEL_PILES = 0


def set_max_parallel_piles(max_parallel_piles: int) -> None:
    """
    Set the (maximal) number of piles to compute in parallel.

    By default, we use all the available hardware threads. Override this by setting the
    ``METACELLS_MAX_PARALLEL_PILES`` environment variable or by invoking this function from the main
    thread.

    A value of ``0`` will use all the available processors (see
    :py:func:`metacells.utilities.parallel.set_processors_count`). Otherwise, the value is the
    positive maximal number of processors to use in parallel for computing piles.

    It may be useful to restrict the number of parallel piles to restrict the total amount of memory
    used by the application, to keep it within the physical RAM available.
    """
    global MAX_PARALLEL_PILES
    MAX_PARALLEL_PILES = max_parallel_piles


set_max_parallel_piles(int(os.environ.get("METACELLS_MAX_PARALLEL_PILES", "0")))


def get_max_parallel_piles() -> int:
    """
    Return the maximal number of piles to compute in parallel.
    """
    return MAX_PARALLEL_PILES


@ut.expand_doc(percent=round((1 + pr.max_gbs) * 100))
def guess_max_parallel_piles(
    cells: AnnData,
    what: str = "__x__",
    *,
    max_gbs: float = pr.max_gbs,
    target_metacell_umis: int = pr.target_metacell_umis,
    cell_umis: Optional[ut.NumpyVector] = pr.cell_umis,
    target_metacell_size: int = pr.target_metacell_size,
    min_target_pile_size: int = pr.min_target_pile_size,
    max_target_pile_size: int = pr.max_target_pile_size,
    target_metacells_in_pile: int = pr.target_metacells_in_pile,
) -> int:
    """
    Try and guess a reasonable maximal number of piles to use for computing metacells for the
    specified ``cells`` using at most ``max_gbs`` of memory (default: {max_gbs}, that is
    {percent}% of all the machine has - if zero or negative, is relative to the machines memory).

    The amount of memory used depends on the target pile size, so give this function the same parameters as for
    :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline` so it will use the same automatically
    adjusted target pile size.

    .. note::

        This is only a best-effort guess. A too-low number would slow down the computation by using
        less processors than it could. A too-high number might cause the computation to crash,
        running out of memory. So: use with care, YMMV, keep an eye on memory usage and other
        applications running in parallel to the computation, and apply common sense.

    .. todo::

        Ideally the system would self tune (avoid spawning more parallel processes for computing
        piles when getting close to the memory limit). This is "not easy" to achieve using Python's
        parallel programming APIs.

    **Computation Parameters**

    1. Figure out the target pile size by invoking
       :py:func:`metacells.pipeline.divide_and_conquer.compute_target_pile_size` using the ``target_metacell_umis``
       (default: {target_metacell_umis}), ``cell_umis`` (default: {cell_umis}), ``target_metacell_size`` (default:
       {target_metacell_size}), ``min_target_pile_size`` (default: {min_target_pile_size}), ``max_target_pile_size``
       (default: {max_target_pile_size}) and ``target_metacells_in_pile`` (default: {target_metacells_in_pile}).

    2. Use ``psutil.virtual_memory`` to query the amount of memory in the system, and apply ``max_gbs`` (default:
       {max_gbs}) off the top (e.g., ``-0.1`` means reduce it by 10%).

    3. Guesstimate the number of piles that, if computed in parallel, will consume this amount of memory.
    """
    target_pile_size = compute_target_pile_size(
        cells,
        what,
        target_metacell_umis=target_metacell_umis,
        cell_umis=cell_umis,
        target_metacell_size=target_metacell_size,
        min_target_pile_size=min_target_pile_size,
        max_target_pile_size=max_target_pile_size,
        target_metacells_in_pile=target_metacells_in_pile,
    )

    cells_nnz = ut.nnz_matrix(ut.get_vo_proper(cells, what))
    piles_nnz = cells_nnz * target_pile_size / cells.n_obs
    parallel_processes = ut.get_processors_count()
    if max_gbs <= 0:
        assert max_gbs > -1
        max_gbs += 1
        max_gbs *= psutil.virtual_memory().total / (1024.0 * 1024.0 * 1024.0)
    parallel_piles = int((max_gbs - cells_nnz * 6.5e-8 - parallel_processes / 13) / (piles_nnz * 13e-8))
    return max(parallel_piles, 1)


@ut.logged()
@ut.expand_doc()
def compute_target_pile_size(
    adata: AnnData,
    what: str = "__x__",
    *,
    cell_umis: Optional[ut.NumpyVector] = pr.cell_umis,
    target_metacell_size: int = pr.target_metacell_size,
    target_metacell_umis: int = pr.target_metacell_umis,
    min_target_pile_size: int = pr.min_target_pile_size,
    max_target_pile_size: int = pr.max_target_pile_size,
    target_metacells_in_pile: int = pr.target_metacells_in_pile,
) -> int:
    """
    Compute how many cells should be in each divide-and-conquer pile.

    Larger piles are slower to process (O(N^2)) than smaller piles, but (up to a point) they allow the algorithm to
    produce better results. The ideal pile size is just big enough to allow for a sufficient number of metacells to be
    created; specifically, we'd like the number of metacells in each pile to be roughly the same as the number of cells
    in a metacell, which is natural for the divide and conquer algorithm.

    Note that the target pile size is just a target. It is directly used in the 1st phase of the divide-and-conquer
    algorithm, but the pile sizes of the 2nd phase are determined by clustering metacells together which inevitably
    cause the actual pile sizes to vary around this number.

    **Computation Parameters**

    0. If ``cell_umis`` is not specified, use the sum of the ``what`` data for each cell.

    1. The natural pile size is the product of the ``target_metacell_size`` (default: {target_metacell_size}) and
       the ``target_metacells_in_pile`` (default: {target_metacells_in_pile}).

    2. Also consider the mean number of cells needed to reach the ``target_metacell_umis`` (default:
       {target_metacell_umis}); this times the ``target_metacells_in_pile`` gives us another pile size, and if this one
       is larger than what we computed in the previous step, we use it instead.

    3. Clamp this value to be no lower than ``min_target_pile_size`` (default: {min_target_pile_size}) and no higher
       than ``max_target_pile_size`` (default: {max_target_pile_size}). That is, if you set both to the same value, this
       will force the target pile size value.
    """
    assert 0 < min_target_pile_size <= max_target_pile_size

    pile_size_by_cells = target_metacell_size * target_metacells_in_pile
    ut.log_calc("pile_size_by_cells", pile_size_by_cells)

    if cell_umis is None:
        cell_umis = ut.get_o_numpy(adata, what, sum=True, formatter=ut.sizes_description).astype("float32")
    else:
        assert cell_umis.dtype == "float32"
    assert isinstance(cell_umis, ut.NumpyVector)
    mean_cell_umis = np.mean(cell_umis)
    ut.log_calc("mean_cell_umis", mean_cell_umis)
    mean_metacell_size = target_metacell_umis / mean_cell_umis
    pile_size_by_umis = int(round(mean_metacell_size * target_metacells_in_pile))
    ut.log_calc("pile_size_by_umis", pile_size_by_umis)

    target_pile_size = max(pile_size_by_cells, pile_size_by_umis)
    target_pile_size = max(target_pile_size, min_target_pile_size)
    target_pile_size = min(target_pile_size, max_target_pile_size)
    ut.log_return("target_pile_size", target_pile_size)
    return target_pile_size


@dataclass
class DirectParameters:  # pylint: disable=too-many-instance-attributes
    """
    Parameters for direct computation of metacells in a single pile.
    """

    select_downsample_min_samples: int
    select_downsample_min_cell_quantile: float
    select_downsample_max_cell_quantile: float
    select_min_gene_total: Optional[int]
    select_min_gene_top3: Optional[int]
    select_min_gene_relative_variance: Optional[float]
    select_min_genes: int
    cells_similarity_value_regularization: float
    cells_similarity_log_data: bool
    cells_similarity_method: str
    target_metacell_umis: int
    cell_umis: ut.NumpyVector
    target_metacell_size: int
    min_metacell_size: int
    knn_k: Optional[int]
    knn_k_umis_quantile: float
    min_knn_k: Optional[int]
    knn_balanced_ranks_factor: float
    knn_incoming_degree_factor: float
    knn_outgoing_degree_factor: float
    knn_min_outgoing_degree: int
    min_seed_size_quantile: float
    max_seed_size_quantile: float
    candidates_cooldown_pass: float
    candidates_cooldown_node: float
    candidates_cooldown_phase: float
    knn_k_size_factor: float
    candidates_min_split_size_factor: float
    candidates_max_merge_size_factor: float
    candidates_max_split_min_cut_strength: Optional[float]
    candidates_min_cut_seed_cells: Optional[int]
    must_complete_cover: bool
    deviants_policy: str
    deviants_gap_skip_cells: int
    deviants_min_gene_fold_factor: float
    deviants_min_noisy_gene_fold_factor: float
    deviants_max_gene_fraction: Optional[float]
    deviants_max_cell_fraction: Optional[float]
    deviants_max_gap_cells_count: int
    deviants_max_gap_cells_fraction: float
    dissolve_min_robust_size_factor: Optional[float]
    dissolve_min_convincing_gene_fold_factor: float
    random_seed: int


@dataclass
class DacParameters:  # pylint: disable=too-many-instance-attributes
    """
    Parameters controlling divide-and-conquer algorithm.
    """

    quick_and_dirty: bool
    groups_similarity_log_data: bool
    groups_similarity_method: str
    min_target_pile_size: int
    max_target_pile_size: int
    target_metacells_in_pile: int
    target_pile_size: int
    piles_knn_k_size_factor: float
    piles_min_split_size_factor: float
    piles_min_robust_size_factor: float
    piles_max_merge_size_factor: float
    direct_parameters: DirectParameters


class SubsetResults:
    """
    The results of computing metacells for a subset of the cells.
    """

    def __init__(self, sdata: AnnData) -> None:
        #: The computed metacell index for each cell.
        self.metacell_indices = ut.get_o_numpy(sdata, "metacell", formatter=ut.groups_description).copy()

        #: The mask of the cells which were in dissolved metacells.
        self.dissolved = ut.get_o_numpy(sdata, "dissolved", formatter=ut.groups_description).copy()

        #: The full index of each cell.
        self.full_cell_indices = ut.get_o_numpy(sdata, "__full_cell_index__")

        #: The mask of the genes that were selected to compute metacells by for the subset.
        self.selected_mask = ut.get_v_numpy(sdata, "selected_gene")

    @ut.logged()
    def collect(
        self,
        *,
        adata: AnnData,
        counts: List[int],
        metacells_level: int,
        collected_mask: ut.NumpyVector,
    ) -> None:
        """
        Collect the results of the subset into the full data.
        """
        metacell_of_cells = ut.get_o_numpy(adata, "metacell", formatter=ut.groups_description).copy()
        dissolved_of_cells = ut.get_o_numpy(adata, "dissolved", formatter=ut.groups_description).copy()
        level_of_cells = ut.get_o_numpy(adata, "metacell_level").copy()

        if np.any(collected_mask):
            max_collected = np.max(metacell_of_cells[collected_mask])
        else:
            max_collected = -1
        assert max_collected + 1 == counts[0]

        metacell_of_subset = self.metacell_indices
        dissolved_of_subset = self.dissolved
        ut.log_calc("subset[metacell]", metacell_of_subset, formatter=ut.groups_description)
        ut.log_calc("subset[dissolved]", dissolved_of_subset)
        metacells_count = round(np.max(metacell_of_subset) + 1)
        metacell_of_subset[metacell_of_subset >= 0] += counts[0]
        counts[0] += metacells_count
        metacell_of_cells[self.full_cell_indices] = metacell_of_subset
        dissolved_of_cells[self.full_cell_indices] = dissolved_of_subset
        level_of_cells[self.full_cell_indices] = metacells_level
        ut.set_o_data(adata, "metacell", metacell_of_cells, formatter=ut.groups_description)
        ut.set_o_data(adata, "dissolved", dissolved_of_cells)
        ut.set_o_data(adata, "metacell_level", level_of_cells)

        assert not np.any(collected_mask[self.full_cell_indices])
        collected_mask[self.full_cell_indices] = True

        selected_genes_mask = ut.get_v_numpy(adata, "selected_gene")
        ut.set_v_data(adata, "selected_gene", selected_genes_mask | self.selected_mask)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def divide_and_conquer_pipeline(
    adata: AnnData,
    what: str = "__x__",
    *,
    rare_max_genes: int = pr.rare_max_genes,
    rare_max_gene_cell_fraction: float = pr.rare_max_gene_cell_fraction,
    rare_min_gene_maximum: int = pr.rare_min_gene_maximum,
    rare_genes_similarity_method: str = pr.rare_genes_similarity_method,
    rare_genes_cluster_method: str = pr.rare_genes_cluster_method,
    rare_min_genes_of_modules: int = pr.rare_min_genes_of_modules,
    rare_min_cells_of_modules: int = pr.rare_min_cells_of_modules,
    rare_min_module_correlation: float = pr.rare_min_module_correlation,
    rare_min_related_gene_fold_factor: float = pr.rare_min_related_gene_fold_factor,
    rare_max_related_gene_increase_factor: float = pr.rare_max_related_gene_increase_factor,
    rare_min_cell_module_total: int = pr.rare_min_cell_module_total,
    rare_max_cells_factor_of_random_pile: float = pr.rare_max_cells_factor_of_random_pile,
    rare_deviants_max_cell_fraction: Optional[float] = pr.rare_deviants_max_cell_fraction,
    rare_dissolve_min_robust_size_factor: Optional[float] = pr.rare_dissolve_min_robust_size_factor,
    rare_dissolve_min_convincing_gene_fold_factor: float = pr.dissolve_min_convincing_gene_fold_factor,
    quick_and_dirty: bool = pr.quick_and_dirty,
    select_downsample_min_samples: int = pr.select_downsample_min_samples,
    select_downsample_min_cell_quantile: float = pr.select_downsample_min_cell_quantile,
    select_downsample_max_cell_quantile: float = pr.select_downsample_max_cell_quantile,
    select_min_gene_total: Optional[int] = pr.select_min_gene_total,
    select_min_gene_top3: Optional[int] = pr.select_min_gene_top3,
    select_min_gene_relative_variance: Optional[float] = pr.select_min_gene_relative_variance,
    select_min_genes: int = pr.select_min_genes,
    cells_similarity_value_regularization: float = pr.cells_similarity_value_regularization,
    cells_similarity_log_data: bool = pr.cells_similarity_log_data,
    cells_similarity_method: str = pr.cells_similarity_method,
    groups_similarity_log_data: bool = pr.groups_similarity_log_data,
    groups_similarity_method: str = pr.groups_similarity_method,
    target_metacell_umis: int = pr.target_metacell_umis,
    cell_umis: Optional[ut.NumpyVector] = pr.cell_umis,
    target_metacell_size: int = pr.target_metacell_size,
    min_metacell_size: int = pr.min_metacell_size,
    target_metacells_in_pile: int = pr.target_metacells_in_pile,
    min_target_pile_size: int = pr.min_target_pile_size,
    max_target_pile_size: int = pr.max_target_pile_size,
    piles_knn_k_size_factor: float = pr.piles_knn_k_size_factor,
    piles_min_split_size_factor: float = pr.piles_min_split_size_factor,
    piles_min_robust_size_factor: float = pr.piles_min_robust_size_factor,
    piles_max_merge_size_factor: float = pr.piles_max_merge_size_factor,
    knn_k: Optional[int] = pr.knn_k,
    knn_k_umis_quantile: float = pr.knn_k_umis_quantile,
    min_knn_k: Optional[int] = pr.min_knn_k,
    knn_balanced_ranks_factor: float = pr.knn_balanced_ranks_factor,
    knn_incoming_degree_factor: float = pr.knn_incoming_degree_factor,
    knn_outgoing_degree_factor: float = pr.knn_outgoing_degree_factor,
    knn_min_outgoing_degree: int = pr.knn_min_outgoing_degree,
    min_seed_size_quantile: float = pr.min_seed_size_quantile,
    max_seed_size_quantile: float = pr.max_seed_size_quantile,
    candidates_knn_k_size_factor: float = pr.candidates_knn_k_size_factor,
    candidates_cooldown_pass: float = pr.cooldown_pass,
    candidates_cooldown_node: float = pr.cooldown_node,
    candidates_cooldown_phase: float = pr.cooldown_phase,
    candidates_min_split_size_factor: float = pr.candidates_min_split_size_factor,
    candidates_max_merge_size_factor: float = pr.candidates_max_merge_size_factor,
    candidates_max_split_min_cut_strength: Optional[float] = pr.max_split_min_cut_strength,
    candidates_min_cut_seed_cells: int = pr.min_cut_seed_cells,
    must_complete_cover: bool = False,
    deviants_policy: str = pr.deviants_policy,
    deviants_gap_skip_cells: int = pr.deviants_gap_skip_cells,
    deviants_min_gene_fold_factor: float = pr.deviants_min_gene_fold_factor,
    deviants_min_noisy_gene_fold_factor: float = pr.deviants_min_noisy_gene_fold_factor,
    deviants_max_gene_fraction: Optional[float] = pr.deviants_max_gene_fraction,
    deviants_max_cell_fraction: Optional[float] = pr.deviants_max_cell_fraction,
    deviants_max_gap_cells_count: int = pr.deviants_max_gap_cells_count,
    deviants_max_gap_cells_fraction: float = pr.deviants_max_gap_cells_fraction,
    dissolve_min_robust_size_factor: Optional[float] = pr.dissolve_min_robust_size_factor,
    dissolve_min_convincing_gene_fold_factor: float = pr.dissolve_min_convincing_gene_fold_factor,
    random_seed: int,
) -> None:
    """
    Complete pipeline using divide-and-conquer to compute the metacells for the ``what`` (default: {what}) data.

    If a ``progress_bar`` is active, progress will be reported into the current (slice of) the progress bar. The
    detection of rare gene modules is not counted in the progress bar; instead it is logged at the ``INFO`` level.

    .. note::

        This is applicable to "any size" data. If the data is "small" (O(10,000)), it will revert to using the direct
        metacell computation (but will still by default first look for rare gene modules). If the data is "large" (up to
        O(10,000,000)), this will be much faster and will require much less memory than using the direct approach. The
        current implementation is not optimized for "huge" data (O(1,000,000,000)) - that is, it will work, and keep
        will use a limited amount of memory, but a faster implementation would distribute the computation across
        multiple servers.

    **Input**

    The presumably "clean" annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation annotation
    containing such a matrix.

    **Returns**

    Sets the following annotations in ``adata``:

    Everything set by ``compute_divide_and_conquer_metacells``; and in addition:

    Variable (Gene) Annotations
        ``rare_gene_module_<N>``
            A boolean mask for the genes in the rare gene module ``N``.

        ``rare_gene``
            A boolean mask for the genes in any of the rare gene modules.

    Observations (Cell) Annotations
        ``cells_rare_gene_module``
            The index of the rare gene module each cell expresses the most, or ``-1`` in the common case it does not
            express any rare genes module.

        ``rare_cell``
            A boolean mask for the (few) cells that express a rare gene module.

    **Computation Parameters**

    0. If ``cell_umis`` is not specified, use the sum of the ``what`` data for each cell.

    1. Invoke :py:func:`metacells.pipeline.divide_and_conquer.compute_target_pile_size` using
       ``target_metacell_umis`` (default: {target_metacell_umis}), ``cell_umis`` (default: {cell_umis}),
       ``target_metacell_size`` (default: {target_metacell_size}), ``min_target_pile_size`` (default:
       {min_target_pile_size}), ``max_target_pile_size`` (default: {max_target_pile_size}) and
       ``target_metacells_in_pile`` (default: {target_metacells_in_pile}).

    2. Invoke :py:func:`metacells.tools.rare.find_rare_gene_modules` to isolate cells expressing
       rare gene modules, using the
       ``rare_max_genes`` (default: {rare_max_genes}),
       ``rare_max_gene_cell_fraction`` (default: {rare_max_gene_cell_fraction}),
       ``rare_min_gene_maximum`` (default: {rare_min_gene_maximum}),
       ``rare_genes_similarity_method`` (default: {rare_genes_similarity_method}),
       ``rare_genes_cluster_method`` (default: {rare_genes_cluster_method}),
       ``rare_min_genes_of_modules`` (default: {rare_min_genes_of_modules}),
       ``rare_min_cells_of_modules`` (default: {rare_min_cells_of_modules}),
       ``rare_min_module_correlation`` (default: {rare_min_module_correlation}),
       ``rare_min_related_gene_fold_factor`` (default: {rare_min_related_gene_fold_factor})
       ``rare_max_related_gene_increase_factor`` (default: {rare_max_related_gene_increase_factor})
       ``rare_max_cells_factor_of_random_pile`` (default: {rare_max_cells_factor_of_random_pile})
       and
       ``rare_min_cell_module_total`` (default: {rare_min_cell_module_total}). Use a non-zero
       ``random_seed`` to make this reproducible.

    3. For each detected rare gene module, collect all cells that express the module, and apply
       :py:func:`metacells.pipeline.direct.compute_direct_metacells` to them.

    4. Collect all the outliers from the above together with the rest of the cells (which express no rare gene module)
       and apply the :py:func:`compute_divide_and_conquer_metacells` to the combined result. The annotations for
       rare-but-outlier cells (e.g. ``cell_directs``) will not reflect the work done when computing the rare gene module
       metacells which they were rejected from.
    """
    if cell_umis is None:
        cell_umis = ut.get_o_numpy(adata, what, sum=True, formatter=ut.sizes_description).astype("float32")
    else:
        assert cell_umis.dtype == "float32"
    assert isinstance(cell_umis, ut.NumpyVector)

    target_pile_size = compute_target_pile_size(
        adata,
        what,
        target_metacell_umis=target_metacell_umis,
        cell_umis=cell_umis,
        target_metacell_size=target_metacell_size,
        target_metacells_in_pile=target_metacells_in_pile,
        min_target_pile_size=min_target_pile_size,
        max_target_pile_size=max_target_pile_size,
    )

    dac_parameters = DacParameters(
        quick_and_dirty=quick_and_dirty,
        groups_similarity_log_data=groups_similarity_log_data,
        groups_similarity_method=groups_similarity_method,
        min_target_pile_size=min_target_pile_size,
        max_target_pile_size=max_target_pile_size,
        target_pile_size=target_pile_size,
        target_metacells_in_pile=target_metacells_in_pile,
        piles_knn_k_size_factor=piles_knn_k_size_factor,
        piles_min_split_size_factor=piles_min_split_size_factor,
        piles_min_robust_size_factor=piles_min_robust_size_factor,
        piles_max_merge_size_factor=piles_max_merge_size_factor,
        direct_parameters=DirectParameters(
            select_downsample_min_samples=select_downsample_min_samples,
            select_downsample_min_cell_quantile=select_downsample_min_cell_quantile,
            select_downsample_max_cell_quantile=select_downsample_max_cell_quantile,
            select_min_gene_total=select_min_gene_total,
            select_min_gene_top3=select_min_gene_top3,
            select_min_gene_relative_variance=select_min_gene_relative_variance,
            select_min_genes=select_min_genes,
            cells_similarity_value_regularization=cells_similarity_value_regularization,
            cells_similarity_log_data=cells_similarity_log_data,
            cells_similarity_method=cells_similarity_method,
            target_metacell_umis=target_metacell_umis,
            cell_umis=cell_umis,
            target_metacell_size=target_metacell_size,
            min_metacell_size=min_metacell_size,
            knn_k=knn_k,
            knn_k_umis_quantile=knn_k_umis_quantile,
            min_knn_k=min_knn_k,
            knn_balanced_ranks_factor=knn_balanced_ranks_factor,
            knn_incoming_degree_factor=knn_incoming_degree_factor,
            knn_outgoing_degree_factor=knn_outgoing_degree_factor,
            knn_min_outgoing_degree=knn_min_outgoing_degree,
            min_seed_size_quantile=min_seed_size_quantile,
            max_seed_size_quantile=max_seed_size_quantile,
            candidates_cooldown_pass=candidates_cooldown_pass,
            candidates_cooldown_node=candidates_cooldown_node,
            candidates_cooldown_phase=candidates_cooldown_phase,
            knn_k_size_factor=candidates_knn_k_size_factor,
            candidates_min_split_size_factor=candidates_min_split_size_factor,
            candidates_max_merge_size_factor=candidates_max_merge_size_factor,
            candidates_max_split_min_cut_strength=candidates_max_split_min_cut_strength,
            candidates_min_cut_seed_cells=candidates_min_cut_seed_cells,
            must_complete_cover=must_complete_cover,
            deviants_policy=deviants_policy,
            deviants_gap_skip_cells=deviants_gap_skip_cells,
            deviants_min_gene_fold_factor=deviants_min_gene_fold_factor,
            deviants_min_noisy_gene_fold_factor=deviants_min_noisy_gene_fold_factor,
            deviants_max_gene_fraction=deviants_max_gene_fraction,
            deviants_max_cell_fraction=deviants_max_cell_fraction,
            deviants_max_gap_cells_count=deviants_max_gap_cells_count,
            deviants_max_gap_cells_fraction=deviants_max_gap_cells_fraction,
            dissolve_min_robust_size_factor=dissolve_min_robust_size_factor,
            dissolve_min_convincing_gene_fold_factor=dissolve_min_convincing_gene_fold_factor,
            random_seed=random_seed,
        ),
    )

    counts = [0]
    _initialize_divide_and_conquer_results(adata)

    if ut.has_progress_bar():
        logger = ut.logger()
        log_level = logger.level
        logger.setLevel(logging.INFO)
        logger.info("Detect rare gene modules...")
        logger.setLevel(log_level)

    name = ut.get_name(adata)

    with ut.timed_step(".rare"):
        tl.find_rare_gene_modules(
            adata,
            what,
            max_genes=rare_max_genes,
            max_gene_cell_fraction=rare_max_gene_cell_fraction,
            min_gene_maximum=rare_min_gene_maximum,
            genes_similarity_method=rare_genes_similarity_method,
            genes_cluster_method=rare_genes_cluster_method,
            min_genes_of_modules=rare_min_genes_of_modules,
            min_cells_of_modules=rare_min_cells_of_modules,
            target_metacell_size=target_metacell_size,
            target_pile_size=target_pile_size,
            min_module_correlation=rare_min_module_correlation,
            min_related_gene_fold_factor=rare_min_related_gene_fold_factor,
            max_related_gene_increase_factor=rare_max_related_gene_increase_factor,
            max_cells_factor_of_random_pile=rare_max_cells_factor_of_random_pile,
            min_cell_module_total=rare_min_cell_module_total,
            reproducible=(random_seed != 0),
        )

        rare_module_of_cells = ut.get_o_numpy(adata, "cells_rare_gene_module", formatter=ut.groups_description)
        rare_cells_count = np.sum(rare_module_of_cells >= 0)
        common_cells_fraction: Optional[float] = None
        if rare_cells_count > 0:
            if ut.has_progress_bar():
                rare_cells_fraction: Optional[float] = rare_cells_count / adata.n_obs
                assert rare_cells_fraction is not None  # Stupid mypy
                common_cells_fraction = 1.0 - rare_cells_fraction
            else:
                rare_cells_fraction = None

            with ut.progress_bar_slice(rare_cells_fraction):
                # TODO: This assumes each rare gene module cells fit in a single pile.
                # If/when we move to a truly ludicrous number of cells, this will no longer be true.
                _compute_metacells_of_piles(
                    adata,
                    what,
                    pile_of_cells=rare_module_of_cells,
                    prefix="rare" if name is None else name + ".rare",
                    counts=counts,
                    collected_mask=np.zeros(adata.n_obs, dtype="bool"),
                    metacells_level=0,
                    direct_parameters=replace(
                        dac_parameters.direct_parameters,
                        deviants_max_cell_fraction=rare_deviants_max_cell_fraction,
                        dissolve_min_robust_size_factor=rare_dissolve_min_robust_size_factor,
                        dissolve_min_convincing_gene_fold_factor=rare_dissolve_min_convincing_gene_fold_factor,
                        must_complete_cover=False,
                    ),
                )

    rare_metacell_of_cells = ut.get_o_numpy(adata, "metacell", formatter=ut.groups_description)
    common_cells_mask = rare_metacell_of_cells < 0
    collected_mask = ~common_cells_mask

    with ut.timed_step(".common"):
        with ut.progress_bar_slice(common_cells_fraction):
            _compute_divide_and_conquer_subset(
                adata,
                what,
                prefix="common" if name is None else name + ".common",
                metacells_level=0,
                subset_mask=common_cells_mask,
                collected_mask=collected_mask,
                counts=counts,
                dac_parameters=dac_parameters,
                random_seed=random_seed,
            )

    if rare_cells_count > 0:
        selected_genes = ut.get_v_numpy(adata, "selected_gene")
        rare_genes = ut.get_v_numpy(adata, "rare_gene")
        ut.set_v_data(adata, "selected_gene", selected_genes | rare_genes)

    _finalize_divide_and_conquer_results(adata)
    assert np.all(collected_mask)


@ut.logged()
def compute_divide_and_conquer_metacells(
    adata: AnnData,
    what: str = "__x__",
    *,
    quick_and_dirty: bool = pr.quick_and_dirty,
    select_downsample_min_samples: int = pr.select_downsample_min_samples,
    select_downsample_min_cell_quantile: float = pr.select_downsample_min_cell_quantile,
    select_downsample_max_cell_quantile: float = pr.select_downsample_max_cell_quantile,
    select_min_gene_total: Optional[int] = pr.select_min_gene_total,
    select_min_gene_top3: Optional[int] = pr.select_min_gene_top3,
    select_min_gene_relative_variance: Optional[float] = pr.select_min_gene_relative_variance,
    select_min_genes: int = pr.select_min_genes,
    cells_similarity_value_regularization: float = pr.cells_similarity_value_regularization,
    cells_similarity_log_data: bool = pr.cells_similarity_log_data,
    cells_similarity_method: str = pr.cells_similarity_method,
    groups_similarity_log_data: bool = pr.groups_similarity_log_data,
    groups_similarity_method: str = pr.groups_similarity_method,
    target_metacell_umis: int = pr.target_metacell_umis,
    cell_umis: Optional[ut.NumpyVector] = pr.cell_umis,
    target_metacell_size: int = pr.target_metacell_size,
    min_metacell_size: int = pr.min_metacell_size,
    min_target_pile_size: int = pr.min_target_pile_size,
    max_target_pile_size: int = pr.max_target_pile_size,
    target_metacells_in_pile: int = pr.target_metacells_in_pile,
    piles_knn_k_size_factor: float = pr.piles_knn_k_size_factor,
    piles_min_split_size_factor: float = pr.piles_min_split_size_factor,
    piles_min_robust_size_factor: float = pr.piles_min_robust_size_factor,
    piles_max_merge_size_factor: float = pr.piles_max_merge_size_factor,
    knn_k: Optional[int] = pr.knn_k,
    knn_k_umis_quantile: float = pr.knn_k_umis_quantile,
    min_knn_k: Optional[int] = pr.min_knn_k,
    knn_balanced_ranks_factor: float = pr.knn_balanced_ranks_factor,
    knn_incoming_degree_factor: float = pr.knn_incoming_degree_factor,
    knn_outgoing_degree_factor: float = pr.knn_outgoing_degree_factor,
    knn_min_outgoing_degree: int = pr.knn_min_outgoing_degree,
    min_seed_size_quantile: float = pr.min_seed_size_quantile,
    max_seed_size_quantile: float = pr.max_seed_size_quantile,
    candidates_cooldown_pass: float = pr.cooldown_pass,
    candidates_cooldown_node: float = pr.cooldown_node,
    candidates_cooldown_phase: float = pr.cooldown_phase,
    candidates_knn_k_size_factor: float = pr.candidates_knn_k_size_factor,
    candidates_min_split_size_factor: float = pr.candidates_min_split_size_factor,
    candidates_max_merge_size_factor: float = pr.candidates_max_merge_size_factor,
    candidates_max_split_min_cut_strength: Optional[float] = pr.max_split_min_cut_strength,
    candidates_min_cut_seed_cells: int = pr.min_cut_seed_cells,
    must_complete_cover: bool = False,
    deviants_policy: str = pr.deviants_policy,
    deviants_gap_skip_cells: int = pr.deviants_gap_skip_cells,
    deviants_min_gene_fold_factor: float = pr.deviants_min_gene_fold_factor,
    deviants_min_noisy_gene_fold_factor: float = pr.deviants_min_noisy_gene_fold_factor,
    deviants_max_gene_fraction: Optional[float] = pr.deviants_max_gene_fraction,
    deviants_max_cell_fraction: Optional[float] = pr.deviants_max_cell_fraction,
    deviants_max_gap_cells_count: int = pr.deviants_max_gap_cells_count,
    deviants_max_gap_cells_fraction: float = pr.deviants_max_gap_cells_fraction,
    dissolve_min_robust_size_factor: Optional[float] = pr.dissolve_min_robust_size_factor,
    dissolve_min_convincing_gene_fold_factor: float = pr.dissolve_min_convincing_gene_fold_factor,
    random_seed: int,
    _fdata: Optional[AnnData] = None,
    _counts: Optional[Dict[str, int]] = None,
) -> None:
    """
    Compute metacells for ``what`` (default: {what}) data using the divide-and-conquer method.

    This divides large data to smaller "piles" and directly computes metacells for each, then generates new piles of
    "similar" metacells and directly computes the final metacells from each such pile. Due to this divide-and-conquer
    approach, the total amount of memory required is bounded (by the pile size), and the amount of total CPU grows
    slowly with the problem size, allowing this method to be applied to very large data (millions of cells).

    If a progress bar is active, progress will be reported into the current (slice of) the progress bar.

    .. todo::

        The divide-and-conquer implementation is restricted for using multiple CPUs on a single shared-memory machine.
        At some problem size (probably around a billion cells), it would make sense to modify it to support distributed
        execution of sub-problems on different servers. This would only be an implementation change, rather than an
        algorithmic change.

    **Input**

    The presumably "clean" annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation annotation
    containing such a matrix.

    **Returns**

    Sets the following annotations in ``adata``:

    Observations (Cell) Annotations
        ``metacell``
            The integer index of the final metacell each cell belongs to. The metacells are in no particular order. This
            is ``-1`` for cells not included in any metacell (outliers, excluded).

        ``dissolve``
           A boolean mask of the cells which were in a dissolved metacell (in the last pile they were in).

    **Computation Parameters**

    0. If ``cell_umis`` is not specified, use the sum of the ``what`` data for each cell.

    1. Invoke :py:func:`metacells.pipeline.divide_and_conquer.compute_target_pile_size` using ``target_metacell_umis``
       (default: {target_metacell_umis}), ``cell_umis`` (default: {cell_umis}), ``target_metacell_size`` (default:
       {target_metacell_size}), ``min_target_pile_size`` (default: {min_target_pile_size}), ``max_target_pile_size``
       (default: {max_target_pile_size}) and ``target_metacells_in_pile`` (default: {target_metacells_in_pile}).

    2. If the data is smaller than the target_pile_size times the ``piles_min_split_size_factor`` (default:
       {piles_min_split_size_factor}), then just invoke :py:func:`metacells.pipeline.direct.compute_direct_metacells`
       using the parameters. Otherwise, perform the following steps.

    3. Phase 0 (preliminary): group the cells randomly into equal-sized piles of roughly the ``target_pile_size`` using
       the ``random_seed`` (default: {random_seed}) to allow making this replicable. Otherwise, use the groups of
       metacells computed in the preliminary phase to create the piles of the final phase.

    4. Compute metacells using the piles by invoking :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
       possibly in parallel.

    5. Collect the outliers from all the piles and recursively compute metacells for them. Continue to compute metacells
       for the outliers-of-outliers, etc. until all the outliers fit in a single pile; then compute metacells for this
       outlier-most cells with ``must_complete_cover=True``.

    6. Phase 0 to 1: Recursively run the algorithm grouping the preliminary metacells into groups-of-metacells, using
       ``piles_knn_k_size_factor`` (default: {piles_knn_k_size_factor}),
       ``piles_min_split_size_factor`` (default: {piles_min_split_size_factor}),
       ``piles_min_robust_size_factor`` (default: {piles_min_robust_size_factor}), and
       ``piles_max_merge_size_factor`` (default: {piles_max_merge_size_factor}).

    7. Phase 1 (final): group the cells into piles based on the groups-of-metacells. Compute metacells using the piles,
       by invoking :py:func:`metacells.pipeline.direct.compute_direct_metacells`, possibly in parallel.

    8. Collect the outliers from all the piles and compute metacells for them (possibly in parallel). If
       ``quick_and_dirty``, stop. Otherwise, continue to group all the cells in metacells by grouping the
       outliers-of-outliers similarly to step 5.

    9. Phase 1 to 2: Recursively run the algorithm grouping the outlier metacells into groups-of-metacells, using
       ``piles_knn_k_size_factor`` (default: {piles_knn_k_size_factor}),
       ``piles_min_split_size_factor`` (default: {piles_min_split_size_factor}),
       ``piles_min_robust_size_factor`` (default: {piles_min_robust_size_factor}), and
       ``piles_max_merge_size_factor`` (default: {piles_max_merge_size_factor}).

    10. Phase 2 (outliers): Group the cells into piles based on the groups-of-metacells. Compute metacells using the
        piles, by invoking :py:func:`metacells.pipeline.direct.compute_direct_metacells`, possibly in parallel. Stop
        here and accept all outliers as final outliers (that is, there are no level 2 metacells).
    """
    if cell_umis is None:
        cell_umis = ut.get_o_numpy(adata, what, sum=True, formatter=ut.sizes_description).astype("float32")
    else:
        assert cell_umis.dtype == "float32"
    assert isinstance(cell_umis, ut.NumpyVector)

    target_pile_size = compute_target_pile_size(
        adata=adata,
        target_metacell_umis=target_metacell_umis,
        cell_umis=cell_umis,
        target_metacell_size=target_metacell_size,
        min_target_pile_size=min_target_pile_size,
        max_target_pile_size=max_target_pile_size,
        target_metacells_in_pile=target_metacells_in_pile,
    )

    dac_parameters = DacParameters(
        quick_and_dirty=quick_and_dirty,
        groups_similarity_log_data=groups_similarity_log_data,
        groups_similarity_method=groups_similarity_method,
        min_target_pile_size=min_target_pile_size,
        max_target_pile_size=max_target_pile_size,
        target_pile_size=target_pile_size,
        target_metacells_in_pile=target_metacells_in_pile,
        piles_knn_k_size_factor=piles_knn_k_size_factor,
        piles_min_split_size_factor=piles_min_split_size_factor,
        piles_min_robust_size_factor=piles_min_robust_size_factor,
        piles_max_merge_size_factor=piles_max_merge_size_factor,
        direct_parameters=DirectParameters(
            select_downsample_min_samples=select_downsample_min_samples,
            select_downsample_min_cell_quantile=select_downsample_min_cell_quantile,
            select_downsample_max_cell_quantile=select_downsample_max_cell_quantile,
            select_min_gene_total=select_min_gene_total,
            select_min_gene_top3=select_min_gene_top3,
            select_min_gene_relative_variance=select_min_gene_relative_variance,
            select_min_genes=select_min_genes,
            cells_similarity_value_regularization=cells_similarity_value_regularization,
            cells_similarity_log_data=cells_similarity_log_data,
            cells_similarity_method=cells_similarity_method,
            target_metacell_umis=target_metacell_umis,
            cell_umis=cell_umis,
            target_metacell_size=target_metacell_size,
            min_metacell_size=min_metacell_size,
            knn_k=knn_k,
            knn_k_umis_quantile=knn_k_umis_quantile,
            min_knn_k=min_knn_k,
            knn_balanced_ranks_factor=knn_balanced_ranks_factor,
            knn_incoming_degree_factor=knn_incoming_degree_factor,
            knn_outgoing_degree_factor=knn_outgoing_degree_factor,
            knn_min_outgoing_degree=knn_min_outgoing_degree,
            min_seed_size_quantile=min_seed_size_quantile,
            max_seed_size_quantile=max_seed_size_quantile,
            candidates_cooldown_pass=candidates_cooldown_pass,
            candidates_cooldown_node=candidates_cooldown_node,
            candidates_cooldown_phase=candidates_cooldown_phase,
            knn_k_size_factor=candidates_knn_k_size_factor,
            candidates_min_split_size_factor=candidates_min_split_size_factor,
            candidates_max_merge_size_factor=candidates_max_merge_size_factor,
            candidates_max_split_min_cut_strength=candidates_max_split_min_cut_strength,
            candidates_min_cut_seed_cells=candidates_min_cut_seed_cells,
            must_complete_cover=must_complete_cover,
            deviants_policy=deviants_policy,
            deviants_gap_skip_cells=deviants_gap_skip_cells,
            deviants_min_gene_fold_factor=deviants_min_gene_fold_factor,
            deviants_min_noisy_gene_fold_factor=deviants_min_noisy_gene_fold_factor,
            deviants_max_gene_fraction=deviants_max_gene_fraction,
            deviants_max_cell_fraction=deviants_max_cell_fraction,
            deviants_max_gap_cells_count=deviants_max_gap_cells_count,
            deviants_max_gap_cells_fraction=deviants_max_gap_cells_fraction,
            dissolve_min_robust_size_factor=dissolve_min_robust_size_factor,
            dissolve_min_convincing_gene_fold_factor=dissolve_min_convincing_gene_fold_factor,
            random_seed=random_seed,
        ),
    )

    _initialize_divide_and_conquer_results(adata)

    collected_mask = np.zeros(adata.n_obs, dtype="bool")

    _compute_divide_and_conquer_subset(
        adata,
        what,
        prefix=ut.get_name(adata) or "full",
        subset_mask=np.full(adata.n_obs, True, dtype="bool"),
        collected_mask=collected_mask,
        metacells_level=0,
        counts=[0],
        dac_parameters=dac_parameters,
        random_seed=random_seed,
    )

    _finalize_divide_and_conquer_results(adata)
    assert np.all(collected_mask)


@ut.logged()
def _initialize_divide_and_conquer_results(adata: AnnData) -> None:
    ut.set_v_data(
        adata,
        "selected_gene",
        np.zeros(adata.n_vars, dtype="bool"),
        formatter=lambda _: "* -> False",
    )
    ut.incremental(adata, "v", "selected_gene")

    for name, value, dtype in (
        ("metacell", -1, "int32"),
        ("dissolved", False, "bool"),
        ("metacell_level", -1, "int32"),
    ):
        ut.incremental(adata, "o", name)
        ut.set_o_data(
            adata,
            name,
            np.full(adata.n_obs, value, dtype=dtype),
            formatter=lambda _: f"* -> {value}",  # pylint: disable=cell-var-from-loop
        )


@ut.logged()
def _reset_subset_results(adata: AnnData, subset_cells_mask: ut.NumpyVector) -> None:
    for name, value, formatter in (("metacell", -1, ut.groups_description),):
        values = ut.get_o_numpy(adata, name, formatter=formatter).copy()
        values[subset_cells_mask] = value
        ut.set_o_data(adata, name, values, formatter=formatter)


@ut.logged()
def _finalize_divide_and_conquer_results(adata: AnnData) -> None:
    ut.done_incrementals(adata)


# NOTE: Any change here must be reflected in _divide_and_conquer_times
@ut.logged()
def _compute_divide_and_conquer_subset(
    adata: AnnData,
    what: str,
    *,
    prefix: str,
    subset_mask: ut.NumpyVector,
    collected_mask: ut.NumpyVector,
    metacells_level: int,
    counts: List[int],
    dac_parameters: DacParameters,
    random_seed: int,
) -> None:
    if not np.any(subset_mask):
        return

    preliminary_pile_of_cells = np.full(adata.n_obs, -1, dtype="int32")
    preliminary_pile_of_cells[subset_mask] = ut.random_piles(
        np.sum(subset_mask),
        target_pile_size=dac_parameters.target_pile_size,
        random_seed=dac_parameters.direct_parameters.random_seed,
    )
    preliminary_piles_count = np.max(preliminary_pile_of_cells) + 1

    time_fractions = _times_to_fractions(
        _divide_and_conquer_times(
            dac_parameters=dac_parameters,
            piles_count=preliminary_piles_count,
            prefix=prefix,
            metacells_level=metacells_level,
        )
    )
    if ut.has_progress_bar():
        time_fractions.reverse()
    else:
        time_fractions = [None] * len(time_fractions)  # type: ignore

    must_cover_dac_parameters = replace(
        dac_parameters, direct_parameters=replace(dac_parameters.direct_parameters, must_complete_cover=True)
    )

    _reset_subset_results(adata, subset_mask)

    if preliminary_piles_count == 1:
        ut.log_calc(f"# {prefix}")
        with ut.progress_bar_slice(time_fractions.pop()):
            _compute_metacells_in_levels(
                adata,
                what,
                prefix=prefix,
                subset_mask=subset_mask,
                collected_mask=collected_mask,
                pile_of_cells=preliminary_pile_of_cells,
                counts=counts,
                metacells_level=metacells_level,
                max_metacells_level=1000,
                dac_parameters=dac_parameters,
            )
        return

    ut.log_calc(f"# {prefix}.preliminary")
    with ut.progress_bar_slice(time_fractions.pop()):
        _compute_metacells_in_levels(
            adata,
            what,
            prefix=prefix + ".preliminary",
            subset_mask=subset_mask,
            collected_mask=np.zeros(adata.n_obs, dtype="bool"),
            pile_of_cells=preliminary_pile_of_cells,
            counts=[0],
            metacells_level=metacells_level,
            max_metacells_level=1000,
            dac_parameters=must_cover_dac_parameters,
        )

    ut.log_calc(f"# {prefix}.groups")
    collect_time = time_fractions.pop()
    groups_time = time_fractions.pop()
    if collect_time is None or groups_time is None:
        total_time = None
    else:
        total_time = collect_time + groups_time
        collect_time /= total_time
        groups_time /= total_time

    with ut.progress_bar_slice(total_time):
        final_pile_of_cells = _compute_metacell_groups(
            adata,
            what,
            collect_time=collect_time,
            groups_time=groups_time,
            prefix=prefix + ".groups",
            subset_mask=subset_mask,
            dac_parameters=must_cover_dac_parameters,
            random_seed=random_seed,
        )

    if dac_parameters.quick_and_dirty:
        ut.log_calc(f"# {prefix}.final")
        with ut.progress_bar_slice(time_fractions.pop()):
            _compute_metacells_in_levels(
                adata,
                what,
                prefix=prefix + ".final",
                subset_mask=subset_mask,
                collected_mask=collected_mask,
                pile_of_cells=final_pile_of_cells,
                counts=counts,
                metacells_level=metacells_level,
                max_metacells_level=1000 if dac_parameters.direct_parameters.must_complete_cover else 1,
                dac_parameters=dac_parameters,
            )
            assert np.all(collected_mask[subset_mask])

    else:
        ut.log_calc(f"# {prefix}.final")
        with ut.progress_bar_slice(time_fractions.pop()):
            _compute_metacells_in_levels(
                adata,
                what,
                prefix=prefix + ".final",
                subset_mask=subset_mask,
                collected_mask=collected_mask,
                pile_of_cells=final_pile_of_cells,
                counts=counts,
                metacells_level=metacells_level,
                max_metacells_level=metacells_level,
                dac_parameters=must_cover_dac_parameters,
            )
            assert np.all(collected_mask[subset_mask])

        if metacells_level == 0 or dac_parameters.direct_parameters.must_complete_cover:
            metacell_of_cells = ut.get_o_numpy(adata, "metacell")
            outliers_mask = subset_mask & (metacell_of_cells < 0)
            collected_mask[outliers_mask] = False

            ut.log_calc(f"# {prefix}.outliers")
            with ut.progress_bar_slice(time_fractions.pop()):
                _compute_divide_and_conquer_subset(
                    adata,
                    what,
                    prefix=prefix + ".outliers",
                    subset_mask=outliers_mask,
                    collected_mask=collected_mask,
                    metacells_level=metacells_level + 1,
                    counts=counts,
                    dac_parameters=dac_parameters,
                    random_seed=random_seed,
                )
                assert np.all(collected_mask[subset_mask])


@ut.logged()
@ut.timed_call()
def _compute_metacells_in_levels(
    adata: AnnData,
    what: str,
    *,
    prefix: str,
    subset_mask: ut.NumpyVector,
    collected_mask: ut.NumpyVector,
    pile_of_cells: ut.NumpyVector,
    counts: List[int],
    metacells_level: int,
    max_metacells_level: int,
    dac_parameters: DacParameters,
) -> None:
    assert metacells_level <= max_metacells_level
    piles_count = np.max(pile_of_cells[subset_mask]) + 1
    assert piles_count > 0

    remaining_time_fraction = 1.0 if ut.has_progress_bar() else None
    outliers_fraction = max(dac_parameters.direct_parameters.deviants_max_cell_fraction or 0.0, 0.05)

    was_last = False
    while True:
        if metacells_level + 1 > max_metacells_level:
            is_last = True
        elif piles_count == 1:
            is_last = was_last
            was_last = True
        else:
            is_last = False

        time_fraction = remaining_time_fraction
        if remaining_time_fraction is not None and not is_last:
            time_fraction = remaining_time_fraction * (1.0 - outliers_fraction)
            remaining_time_fraction *= outliers_fraction

        with ut.progress_bar_slice(time_fraction):
            _compute_metacells_of_piles(
                adata,
                what,
                prefix=f"{prefix}.level.{metacells_level}",
                pile_of_cells=pile_of_cells,
                counts=counts,
                collected_mask=collected_mask,
                direct_parameters=dac_parameters.direct_parameters,
                metacells_level=metacells_level + 1,
            )
            assert np.all(collected_mask[subset_mask])

        if is_last:
            return

        metacells_level += 1
        metacell_of_cells = ut.get_o_numpy(adata, "metacell")
        subset_mask = subset_mask & (metacell_of_cells < 0)
        subset_count = int(np.sum(subset_mask))
        if subset_count == 0:
            return

        collected_mask[subset_mask] = False
        pile_of_cells = np.full(adata.n_obs, -1, dtype="int32")
        pile_of_cells[subset_mask] = ut.random_piles(
            subset_count,
            target_pile_size=dac_parameters.target_pile_size,
            random_seed=dac_parameters.direct_parameters.random_seed,
        )


@ut.logged()
@ut.timed_call()
def _compute_metacell_groups(
    adata: AnnData,
    what: str,
    *,
    collect_time: Optional[float],
    groups_time: Optional[float],
    prefix: str,
    subset_mask: ut.NumpyVector,
    dac_parameters: DacParameters,
    random_seed: int,
) -> ut.NumpyVector:
    metacell_of_cells = ut.get_o_numpy(adata, "metacell")
    assert not np.any(metacell_of_cells[subset_mask] < 0)

    with ut.progress_bar_slice(collect_time):
        sdata = ut.slice(
            adata,
            name=f"{prefix}.grouped",
            top_level=False,
            obs=subset_mask,
            track_obs="__full_cell_index__",
        )

        mdata = collect_metacells(
            sdata,
            what,
            groups=metacell_of_cells[subset_mask],
            name=prefix,
            _metacell_groups=True,
            top_level=False,
            random_seed=random_seed,
        )

    with ut.progress_bar_slice(groups_time):
        metacell_sizes = ut.get_o_numpy(mdata, "grouped").astype("float32")
        target_pile_size = compute_target_pile_size(
            mdata,
            what,
            target_metacell_umis=dac_parameters.target_pile_size,
            cell_umis=metacell_sizes,
            target_metacell_size=dac_parameters.target_metacells_in_pile,
            min_target_pile_size=dac_parameters.min_target_pile_size,
            max_target_pile_size=dac_parameters.max_target_pile_size,
            target_metacells_in_pile=dac_parameters.target_metacells_in_pile,
        )

        group_dac_parameters = replace(
            dac_parameters,
            target_pile_size=target_pile_size,
            direct_parameters=replace(
                dac_parameters.direct_parameters,
                target_metacell_umis=dac_parameters.target_pile_size,
                cell_umis=metacell_sizes,
                target_metacell_size=dac_parameters.target_metacells_in_pile,
                knn_k_size_factor=dac_parameters.piles_knn_k_size_factor,
                candidates_min_split_size_factor=dac_parameters.piles_min_split_size_factor,
                candidates_max_merge_size_factor=dac_parameters.piles_max_merge_size_factor,
                cells_similarity_log_data=dac_parameters.groups_similarity_log_data,
                cells_similarity_method=dac_parameters.groups_similarity_method,
                dissolve_min_robust_size_factor=dac_parameters.piles_min_robust_size_factor,
                must_complete_cover=True,
            ),
        )

        _initialize_divide_and_conquer_results(mdata)
        collected_mask = np.zeros(mdata.n_obs, dtype="bool")

        _compute_divide_and_conquer_subset(
            mdata,
            what,
            prefix=prefix,
            subset_mask=np.full(mdata.n_obs, True, dtype="bool"),
            collected_mask=collected_mask,
            metacells_level=0,
            counts=[0],
            dac_parameters=group_dac_parameters,
            random_seed=random_seed,
        )

        _finalize_divide_and_conquer_results(mdata)
        assert np.all(collected_mask)

        pile_of_metacells = ut.get_o_numpy(mdata, "metacell", formatter=ut.groups_description)
        assert not np.any(pile_of_metacells < 0)

        pile_of_cells = np.full(adata.n_obs, -1, dtype="int32")
        pile_of_cells[subset_mask] = pile_of_metacells[metacell_of_cells[subset_mask]]

    groups_selected_genes_mask = ut.get_v_numpy(mdata, "selected_gene")
    metacells_selected_genes_mask = ut.get_v_numpy(adata, "selected_gene")
    ut.set_v_data(adata, "selected_gene", groups_selected_genes_mask | metacells_selected_genes_mask)

    return pile_of_cells


@ut.logged()
@ut.timed_call()
def _compute_metacells_of_piles(
    adata: AnnData,
    what: str,
    *,
    prefix: str,
    pile_of_cells: ut.NumpyVector,
    counts: List[int],
    collected_mask: ut.NumpyVector,
    direct_parameters: DirectParameters,
    metacells_level: int,
) -> int:
    piles_count = np.max(pile_of_cells) + 1

    @ut.timed_call()
    def _compute_pile_metacells(pile_index: int) -> SubsetResults:
        pile_cells_mask = pile_of_cells == pile_index
        assert np.any(pile_cells_mask)
        pdata = ut.slice(
            adata,
            name=f"{prefix}.{pile_index}/{piles_count}" if piles_count > 1 else prefix,
            top_level=False,
            obs=pile_cells_mask,
            track_obs="__full_cell_index__",
        )
        parameters = direct_parameters.__dict__.copy()
        parameters["cell_umis"] = direct_parameters.cell_umis[pile_cells_mask]
        if piles_count > 1:
            parameters["must_complete_cover"] = False
        with ut.timed_step(".direct"):
            compute_direct_metacells(pdata, what, **parameters)
        subset_results = SubsetResults(pdata)
        pdata = None
        gc.collect()
        return subset_results

    with ut.timed_step(".prepare"):
        ut.get_vo_proper(adata, what, layout="row_major")

    with ut.timed_step(".compute"):
        gc.collect()
        ut.logger().debug("MAX_PARALLEL_PILES: %s", get_max_parallel_piles())
        subset_results_of_piles = ut.parallel_map(
            _compute_pile_metacells,
            piles_count,
            max_processors=get_max_parallel_piles(),
        )

    with ut.timed_step(".collect"):
        for pile_index, subset_results in enumerate(subset_results_of_piles):
            with ut.log_step(
                "- pile",
                pile_index,
                formatter=lambda pile_index: ut.progress_description(piles_count, pile_index, "pile"),
            ):
                subset_results.collect(
                    adata=adata,
                    counts=counts,
                    collected_mask=collected_mask,
                    metacells_level=metacells_level,
                )

    ut.log_calc("collected counts", counts)
    return piles_count


# NOTE: Any change here must be reflected in  _compute_divide_and_conquer_subset
def _divide_and_conquer_times(
    *,
    prefix: str,
    piles_count: int,
    metacells_level: int,
    dac_parameters: DacParameters,
) -> List[float]:
    if piles_count == 1:
        ut.log_calc(f"expected {prefix} single pile time", 1.0)
        return [1.0]

    must_cover_dac_parameters = replace(
        dac_parameters, direct_parameters=replace(dac_parameters.direct_parameters, must_complete_cover=True)
    )

    times = [
        np.sum(
            _levels_times(
                prefix=prefix + ".preliminary",
                piles_count=piles_count,
                metacells_level=metacells_level,
                max_metacells_level=1000,
                dac_parameters=must_cover_dac_parameters,
            )
        ),
        1.0,
        np.sum(
            _group_times(
                prefix=prefix + ".groups",
                piles_count=piles_count,
                dac_parameters=must_cover_dac_parameters,
            )
        ),
    ]

    if dac_parameters.quick_and_dirty:
        times += [
            _levels_times(
                prefix=prefix + ".final",
                piles_count=piles_count,
                metacells_level=metacells_level,
                max_metacells_level=1000
                if dac_parameters.direct_parameters.must_complete_cover
                else metacells_level + 1,
                dac_parameters=dac_parameters,
            )
        ]

    else:
        times += _levels_times(
            prefix=prefix + ".final",
            piles_count=piles_count,
            metacells_level=metacells_level,
            max_metacells_level=metacells_level,
            dac_parameters=dac_parameters,
        )

        if metacells_level == 0 or dac_parameters.direct_parameters.must_complete_cover:
            outliers_fraction = max(dac_parameters.direct_parameters.deviants_max_cell_fraction or 0.0, 0.05)
            outlier_piles_count = max(1, round(piles_count * outliers_fraction))
            times += [
                np.sum(
                    _divide_and_conquer_times(
                        prefix=prefix + ".outliers",
                        piles_count=outlier_piles_count,
                        metacells_level=metacells_level + 1,
                        dac_parameters=dac_parameters,
                    )
                )
            ]

    return times


def _levels_times(
    *,
    prefix: str,
    piles_count: int,
    metacells_level: int,
    max_metacells_level: int,
    dac_parameters: DacParameters,
) -> List[float]:
    assert metacells_level <= max_metacells_level
    assert piles_count > 1

    times: List[float] = []

    while True:
        if metacells_level > max_metacells_level:
            return times

        if piles_count == 1:
            ut.log_calc(f"expected {prefix}.level.{metacells_level} single pile time", 1.0)
            times.append(1.0)
            return times

        piles_time = _piles_time(piles_count)
        ut.log_calc(f"expected {prefix}.level.{metacells_level} {piles_count} piles time", piles_time)
        times.append(piles_time)

        outliers_fraction = max(dac_parameters.direct_parameters.deviants_max_cell_fraction or 0.0, 0.05)
        piles_count = max(1, round(piles_count * outliers_fraction))
        metacells_level += 1


def _group_times(
    *,
    prefix: str,
    piles_count: int,
    dac_parameters: DacParameters,
) -> List[float]:
    assert dac_parameters.direct_parameters.must_complete_cover
    expected_mean_pre_metacells_per_group = dac_parameters.target_metacells_in_pile
    expected_groups_target_pile_size = round(
        expected_mean_pre_metacells_per_group * dac_parameters.target_metacells_in_pile
    )
    expected_groups_target_pile_size = max(expected_groups_target_pile_size, dac_parameters.min_target_pile_size)
    expected_groups_target_pile_size = min(expected_groups_target_pile_size, dac_parameters.max_target_pile_size)
    expected_total_pre_metacells = piles_count * dac_parameters.target_metacells_in_pile
    expected_groups_piles_count = max(1, round(expected_total_pre_metacells / expected_groups_target_pile_size))

    return _divide_and_conquer_times(
        prefix=prefix,
        piles_count=expected_groups_piles_count,
        metacells_level=0,
        dac_parameters=dac_parameters,
    )


def _piles_time(piles_count: int) -> float:
    parallel_processors = ut.get_processors_count()
    if MAX_PARALLEL_PILES > 0:
        parallel_processors = min(parallel_processors, MAX_PARALLEL_PILES)
    return ceil(piles_count / min(parallel_processors, piles_count))


def _times_to_fractions(times: List[float]) -> List[float]:
    total_time = sum(times)
    return [time / total_time for time in times]



"""
Typing
------

The code has to deal with many different alternative data types for what is essentially two basic
data types: 2D matrices and 1D vectors.

Specifically, we have pandas data frames and series, Scipy sparse matrices, and numpy
multi-dimensional arrays (not to mention the deprecated numpy matrix type).

Python has the great ability to "duck type", so in an ideal world, we could just pretend these are
just two data types and be done. In practice, this is hopelessly broken.

First, even operations that exists for all data types sometimes have different interfaces
(as in, ``np.foo(matrix, ...)`` vs. ``matrix.foo(...)``).

Second, operating on sparse and dense data often requires completely different code paths.

This makes it very easy to write code that works today and breaks tomorrow when someone passes a
pandas series to a function that expects a numpy array and it just *almost* works correctly (and god
help the poor soul that mixes up a numpy matrix with a numpy 2d array, or passes a categorical
pandas series to something that expects a series of strings).

"Eternal vigilance is the cost of freedom" - the solution here is to define a bunch of fake types,
which are almost entirely for the benefit of the ``mypy`` type checker (with some run-time
assertions as well).

This not only makes the code intent explicit ("explicit is better than implicit") but also allows us
to leverage ``mypy`` to catch errors such as applying a numpy operation on a sparse matrix, etc.

To put some order in this chaos, the following concepts are used:

* :py:const:`Shaped` is any 1d or 2d data in any format we can work with. :py:const:`Matrix` is
  any 2d data, and :py:const:`Vector` is any 1d data.

* For 2D data, we allow multiple data types that we can't directly operate on:
  most :py:class:`SparseMatrix` layouts, :py:class:`PandasFrame` and ``np.matrix`` have
  strange quirks when it comes to directly operating on them and should be avoided, while CSR and
  CSC :py:class:`CompressedMatrix` sparse matrices and properly-laid-out 2D numpy arrays
  :py:const:`NumpyMatrix` are in general well-behaved. We therefore introduce the concept of
  :py:const:`ProperMatrix` vs. :py:const:`ImproperMatrix` types, and provide functions that
  manipulate whether the "proper" data is in row-major or column-major order.

* For 1D data, we just distinguish between :py:class:`PandasSeries` and 1D numpy
  :py:const:`NumpyVector` arrays, as these are the only types we allow. In theory we could have also
  allowed for sparse vectors but mercifully these are very uncommon so we can just ignore them.

Ironically, now that ``numpy`` added type annotations, the usefulness of the type hints added here
has decreased, since both :py:const:`NumpyVector` and :py:const:`NumpyMatrix` are aliases to the
same ``numpy.ndaarray`` type. Perhaps in the future numpy would allow for using ``Annotated`` types
(with explicit number of dimensions, or even - gasp - the element data type) to allow for more
useful type annotations. Or this could all be ported to Julia and avoid this whole mess.
"""

from abc import abstractmethod
from contextlib import contextmanager
from typing import Any
from typing import Collection
from typing import Iterator
from typing import Optional
from typing import Sized
from typing import Tuple
from typing import TypeVar
from typing import Union

import numpy as np
import pandas as pd  # type: ignore
import scipy.sparse as sp  # type: ignore

import metacells.utilities.documentation as utd
import metacells.utilities.timing as utm

try:
    from typing_extensions import Protocol
except ModuleNotFoundError:

    class Protocol:  # type: ignore
        """
        Placeholder if we don't have ``typing_extensions``.
        """


__all__ = [
    "CPP_DATA_TYPES",
    "Shaped",
    "ProperShaped",
    "ImproperShaped",
    "Matrix",
    "ProperMatrix",
    "NumpyMatrix",
    "CompressedMatrix",
    "ImproperMatrix",
    "SparseMatrix",
    "PandasFrame",
    "Vector",
    "NumpyVector",
    "ImproperVector",
    "PandasSeries",
    "is_1d",
    "is_2d",
    "maybe_numpy_vector",
    "maybe_numpy_matrix",
    "maybe_sparse_matrix",
    "maybe_compressed_matrix",
    "maybe_pandas_frame",
    "maybe_pandas_series",
    "mustbe_numpy_vector",
    "mustbe_numpy_matrix",
    "mustbe_sparse_matrix",
    "mustbe_compressed_matrix",
    "mustbe_pandas_frame",
    "mustbe_pandas_series",
    "to_proper_matrix",
    "to_proper_matrices",
    "to_pandas_series",
    "to_pandas_frame",
    "frozen",
    "freeze",
    "unfreeze",
    "unfrozen",
    "to_numpy_matrix",
    "to_numpy_vector",
    "DENSE_FAST_FLAG",
    "SPARSE_FAST_FORMAT",
    "SPARSE_SLOW_FORMAT",
    "LAYOUT_OF_AXIS",
    "PER_OF_AXIS",
    "shaped_dtype",
    "matrix_layout",
    "is_layout",
    "is_contiguous",
    "to_contiguous",
    "mustbe_canonical",
    "is_canonical",
    "eliminate_zeros",
    "sort_indices",
    "sum_duplicates",
    "shaped_checksum",
]


#: Numpy 2-dimensional data.
#:
#: .. note::
#:
#:    This is not to be confused with ``numpy.matrix`` which must not be used, but is returned
#:    by the occasional function, and would wreak havoc on the semantics of some operations unless
#:    immediately concerted to a proper ``NumpyMatrix``, which is a simple 2-dimensional
#:    ``ndarray``.
NumpyMatrix = np.ndarray  # Should be: Annotated[np.ndarray, NDim(2)]

#: Numpy 1-dimensional data.
NumpyVector = np.ndarray  # Should be: Annotated[np.ndarray, NDim(1)]


# pylint: disable=missing-function-docstring


S = TypeVar("S", bound="ShapedProtocol")


class ShapedProtocol(Protocol):
    """
    A ``mypy`` type for any shaped (1- or 2-dimensional, proper or improper) data.
    """

    ndim: int
    shape: Union[Tuple[int, int], Tuple[int]]

    def __getitem__(self, key: Any) -> Any:
        ...

    def __setitem__(self, key: Any, value: Any) -> Any:
        ...

    def transpose(self: S) -> S:
        ...


SP = TypeVar("SP", bound="SparseMatrix")


class SparseMatrix(ShapedProtocol, Protocol):
    """
    A ``mypy`` type for sparse 2-dimensional data.

    Should have been ``SparseMatrix = sp.base.spmatrix``.
    """

    shape: Tuple[int, int]
    nnz: int

    def getformat(self) -> str:
        ...

    def toarray(self, order: Optional[str] = None) -> NumpyMatrix:
        ...

    def multiply(self: SP, other: ShapedProtocol) -> SP:
        ...

    def getcol(self: SP, index: int) -> SP:
        ...

    def getrow(self: SP, index: int) -> SP:
        ...

    def sum(self, *, axis: int) -> NumpyVector:
        ...

    def max(self, *, axis: int) -> NumpyVector:
        ...

    def nanmax(self, *, axis: int) -> NumpyVector:
        ...

    def min(self, *, axis: int) -> NumpyVector:
        ...

    def nanmin(self, *, axis: int) -> NumpyVector:
        ...

    def getnnz(self, *, axis: int) -> NumpyVector:
        ...

    def argmax(self, *, axis: int) -> NumpyVector:
        ...

    def maximum(self: SP, other: SP) -> SP:
        ...

    def tocsr(self) -> "CompressedMatrix":
        ...

    def tocsc(self) -> "CompressedMatrix":
        ...

    def nonzero(self) -> Tuple[np.ndarray, np.ndarray]:
        ...

    def mean(self: SP, *, axis: int) -> "np.ndarray":
        ...

    def copy(self: SP) -> SP:
        ...


class CompressedMatrix(SparseMatrix, Protocol):
    """
    A ``mypy`` type for sparse CSR/CSC 2-dimensional data.

    Should have been ``CompressedMatrix = sp..._cs_matrix``.
    """

    indices: np.ndarray
    indptr: np.ndarray
    data: np.ndarray
    has_sorted_indices: bool
    has_canonical_format: bool

    def sum_duplicates(self) -> None:
        ...

    def eliminate_zeros(self) -> None:
        ...

    def sort_indices(self) -> None:
        ...


class PandasIndex(ShapedProtocol, Collection, Sized, Protocol):
    """
    A ``mypy`` type for a pandas index.
    """

    values: NumpyVector


class PandasFrame(ShapedProtocol, Protocol):
    """
    A ``mypy`` type for pandas 2-dimensional data.

    Should have been ``PandasFrame = pd.DataFrame``.
    """

    shape: Tuple[int, int]
    values: NumpyMatrix
    index: PandasIndex
    columns: PandasIndex

    def __delitem__(self, key: Any) -> None:
        ...

    def __constraints__(self, key: Any) -> bool:
        ...


class PandasSeries(ShapedProtocol, Sized, Protocol):
    """
    A ``mypy`` type for pandas 1-dimensional data.

    Should have been ``PandasSeries = pd.Series``.
    """

    size: int
    shape: Tuple[int]
    values: NumpyVector
    index: PandasIndex

    @abstractmethod
    def sort_values(self, inplace: bool, ascending: bool) -> None:
        ...


# pylint: enable=missing-function-docstring


#: The data types supported by the C++ extensions code.
CPP_DATA_TYPES = ["float32", "float64", "int32", "int64", "uint32", "uint64"]

#: A ``mypy`` type for "proper" 2-dimensional data.
#:
#: "Proper" data allows for direct processing without having
#: to mess with its formatting.
ProperMatrix = Union[NumpyMatrix, CompressedMatrix]

#: A ``mypy`` type for "improper" 2-dimensional data.
#:
#: "Improper" data contains or can be converted to "proper" data.
ImproperMatrix = Union[PandasFrame, SparseMatrix]

#: A ``mypy`` type for any 2-dimensional data.
Matrix = Union[ProperMatrix, ImproperMatrix]

#: An "improper" 1-dimensional data.
ImproperVector = Union[Collection[int], Collection[float], PandasSeries]

#: A ``mypy`` type for any 1-dimensional data.
#:
#: .. todo::
#:
#:    Is there any need for ``SparseVector``?
Vector = Union[NumpyVector, ImproperVector]

#: A "proper" 1- or 2-dimensional data.
ProperShaped = Union[ProperMatrix, NumpyVector]

#: An "improper" 1- or 2- dimensional data.
ImproperShaped = Union[ImproperMatrix, ImproperVector]

#: Shaped data of any of the types we can deal with.
Shaped = Union[ProperShaped, ImproperShaped]

#: Pandas data in various types.
PandasShaped = Union[PandasFrame, PandasSeries]


def is_1d(shaped: Shaped) -> bool:
    """
    Test whether the ``shaped`` is 1-dimensional.
    """
    return hasattr(shaped, "ndim") and getattr(shaped, "ndim") == 1


def is_2d(shaped: Shaped) -> bool:
    """
    Test whether the ``shaped`` is 2-dimensional.
    """
    return hasattr(shaped, "ndim") and getattr(shaped, "ndim") == 2


def maybe_numpy_vector(shaped: Any) -> Optional[NumpyVector]:
    """
    Return the ``shaped`` as a :py:const:`NumpyVector`, if it is one.
    """
    if isinstance(shaped, np.ndarray) and shaped.ndim == 1:
        return shaped
    return None


def maybe_numpy_matrix(shaped: Any) -> Optional[NumpyMatrix]:
    """
    Return the ``shaped`` as a :py:const:`NumpyMatrix`, if it is one.

    .. note::

        This looks for a 2-dimensional ``numpy.ndarray`` which is **not** a ``numpy.matrix``. Do not
        use ``numpy.matrix`` - it is deprecated and behaves subtly different to a 2-dimensional
        ``numpy.ndarray`` leading to hard-to-find bugs.
    """
    if isinstance(shaped, np.ndarray) and shaped.ndim == 2 and not isinstance(shaped, np.matrix):
        return shaped
    return None


def maybe_sparse_matrix(shaped: Any) -> Optional[SparseMatrix]:
    """
    Return ``shap`` as a :py:const:`SparseMatrix`, if it is one.

    .. note::

        This will succeed for a :py:const:`CompressedMatrix` which is a sub-type of a
        :py:const:`SparseMatrix`.
    """
    if isinstance(shaped, sp.base.spmatrix):
        return shaped
    return None


def maybe_compressed_matrix(shaped: Any) -> Optional[CompressedMatrix]:
    """
    Return ``shaped`` as a :py:const:`CompressedMatrix`, if it is one.
    """
    if isinstance(shaped, (sp.csr_matrix, sp.csc_matrix)):
        return shaped
    return None


def maybe_pandas_series(shaped: Any) -> Optional[PandasSeries]:
    """
    Return ``shaped`` as a :py:const:`PandasSeries`, if it is one.
    """
    if isinstance(shaped, pd.Series):
        return shaped
    return None


def maybe_pandas_frame(shaped: Any) -> Optional[PandasFrame]:
    """
    Return ``shaped`` s a :py:const:`PandasFrame`, if it is one.
    """
    if isinstance(shaped, pd.DataFrame):
        return shaped
    return None


def mustbe_numpy_vector(shaped: Any) -> NumpyVector:
    """
    Return ``shaped`` as a :py:const:`NumpyVector`, asserting it must be one.
    """
    assert isinstance(shaped, np.ndarray) and shaped.ndim == 1
    return shaped


def mustbe_numpy_matrix(shaped: Any) -> NumpyMatrix:
    """
    Return ``shaped`` as a :py:const:`NumpyMatrix`, asserting it must be one.

    .. note::

        This looks for a 2-dimensional ``numpy.ndarray`` which is **not** a ``numpy.matrix``. Do not
        use ``numpy.matrix`` - it is deprecated and behaves subtly different to a 2-dimensional
        ``numpy.ndarray`` leading to hard-to-find bugs.
    """
    assert isinstance(shaped, np.ndarray) and shaped.ndim == 2 and not isinstance(shaped, np.matrix)
    return shaped


def mustbe_sparse_matrix(shaped: Any) -> SparseMatrix:
    """
    Return ``shaped`` as a :py:const:`SparseMatrix`, asserting it must be one.

    .. note::

        This will succeed for a :py:const:`CompressedMatrix` which is a sub-type of a
        :py:const:`SparseMatrix`.
    """
    assert isinstance(shaped, sp.base.spmatrix)
    return shaped


def mustbe_compressed_matrix(shaped: Any) -> CompressedMatrix:
    """
    Return ``shaped`` as a :py:const:`CompressedMatrix`, asserting it must be one.
    """
    assert isinstance(shaped, (sp.csr_matrix, sp.csc_matrix))
    return shaped


def mustbe_pandas_series(shaped: Any) -> PandasSeries:
    """
    Return ``shaped`` as a :py:const:`PandasSeries`, asserting it must be one.
    """
    assert isinstance(shaped, pd.Series)
    return shaped


def mustbe_pandas_frame(shaped: Any) -> PandasFrame:
    """
    Return ``shaped`` as a :py:const:`PandasFrame`, asserting it must be one.
    """
    assert isinstance(shaped, pd.DataFrame)
    return shaped


@utd.expand_doc()
def to_proper_matrix(matrix: Matrix, *, default_layout: str = "row_major") -> ProperMatrix:
    """
    Given some 2D ``matrix``, return in in a :py:const:`ProperMatrix` format we can safely process.

    If the data is in some strange sparse format, use ``default_layout`` (default: {default_layout})
    to decide whether to return it in ``row_major`` (CSR) or ``column_major`` (CSC) layout.
    """
    if matrix.ndim != 2:
        raise ValueError(f"data is {matrix.ndim}-dimensional, " "expected 2-dimensional")

    if default_layout not in LAYOUT_OF_AXIS:
        raise ValueError(f"invalid default layout: {default_layout}")

    frame = maybe_pandas_frame(matrix)
    if frame is not None:
        matrix = frame.values
        if isinstance(matrix, pd.core.arrays.categorical.Categorical):
            matrix = np.array(matrix)

    compressed = maybe_compressed_matrix(matrix)
    if compressed is not None:
        return compressed

    sparse = maybe_sparse_matrix(matrix)
    if sparse is not None:
        if default_layout == "column_major":
            with utm.timed_step("matrix.tocsc"):
                utm.timed_parameters(results=sparse.shape[1], elements=sparse.nnz / sparse.shape[1])
                return sparse.tocsc()

        with utm.timed_step("matrix.tocsr"):
            utm.timed_parameters(results=sparse.shape[0], elements=sparse.nnz / sparse.shape[0])
            return sparse.tocsr()

    dense = maybe_numpy_matrix(matrix)
    if dense is None:
        dense = np.asarray(matrix)

    return dense


def to_proper_matrices(
    matrix: Matrix, *, default_layout: str = "row_major"
) -> Tuple[ProperMatrix, Optional[NumpyMatrix], Optional[CompressedMatrix]]:
    """
    Similar to :py:func:`to_proper_matrix` but return a tuple with the proper matrix and also its
    :py:const:`NumpyMatrix` representation and its py:const:`CompressedMatrix` representation.
    Exactly one of these two representations will be ``None``.

    If the data is in some strange sparse format, use ``default_layout`` (default: {default_layout})
    to decide whether to return it in ``row_major`` (CSR) or ``column_major`` (CSC) layout.

    This is used to pick between dense and compressed code paths, and provides typed references so
    ``mypy`` can type-check each of these paths:

    .. code:: python

        proper, dense, compressed = to_proper_matrices(matrix)

        ... Common code path can use the proper matrix value ...

        if dense is not None:
            assert compressed is None
            ... Dense code path can use the dense matrix ...

        else:
            assert compressed is not None
            ... Compressed code path can use the compressed matrix value ...

            if metacells.ut.matrix_layout(compressed) == 'row_major':
                ... CSR code path ...
            else:
                ... CSC code path ...
    """
    proper = to_proper_matrix(matrix, default_layout=default_layout)
    dense = maybe_numpy_matrix(proper)
    compressed = maybe_compressed_matrix(proper)
    assert (dense is None) or (compressed is None)
    assert (dense is None) != (compressed is None)
    return (proper, dense, compressed)


def to_pandas_series(
    vector: Optional[Vector] = None,
    *,
    index: Optional[Vector] = None,
) -> PandasSeries:
    """
    Construct a pandas series from any :py:const:`Vector`.
    """
    if vector is None:
        return pd.Series(index=index)

    return pd.Series(to_numpy_vector(vector), index=index)


def to_pandas_frame(
    matrix: Optional[Matrix] = None, *, index: Optional[Vector] = None, columns: Optional[Vector] = None
) -> PandasFrame:
    """
    Construct a pandas frame from any :py:const:`Matrix`.
    """
    if matrix is None:
        return pd.DataFrame(index=index, columns=columns)

    sparse = maybe_sparse_matrix(matrix)
    if sparse is not None:
        return pd.DataFrame.sparse.from_spmatrix(sparse, index=index, columns=columns)

    return pd.DataFrame(to_numpy_matrix(matrix, only_extract=True), index=index, columns=columns)


def frozen(shaped: Union[ProperShaped, PandasShaped]) -> bool:
    """
    Test whether the ``shaped`` data is protected against future modification.
    """
    compressed = maybe_compressed_matrix(shaped)
    if compressed is not None:
        assert (
            compressed.indices.flags.writeable == compressed.indptr.flags.writeable == compressed.data.flags.writeable
        )
        return not compressed.data.flags.writeable

    if isinstance(shaped, (pd.DataFrame, pd.Series, pd.core.indexes.base.Index)):
        shaped = shaped.values

    if isinstance(shaped, pd.core.arrays.categorical.Categorical):
        shaped = shaped.codes

    if isinstance(shaped, np.ndarray):
        return not shaped.flags.writeable

    raise NotImplementedError(f"frozen of {shaped.__class__}")


def freeze(shaped: Union[ProperShaped, PandasShaped]) -> None:
    """
    Protect the ``shaped`` data against future modification.
    """
    compressed = maybe_compressed_matrix(shaped)
    if compressed is not None:
        compressed.indices.setflags(write=False)
        compressed.indptr.setflags(write=False)
        compressed.data.setflags(write=False)
        return

    if isinstance(shaped, (pd.DataFrame, pd.Series, pd.core.indexes.base.Index)):
        shaped = shaped.values

    if isinstance(shaped, pd.core.arrays.categorical.Categorical):
        shaped = shaped.codes

    if isinstance(shaped, np.ndarray):
        shaped.setflags(write=False)
        return

    raise NotImplementedError(f"freeze of {shaped.__class__}")


def unfreeze(shaped: Union[ProperShaped, PandasShaped]) -> None:
    """
    Permit future modification of some ``shaped`` data.
    """
    compressed = maybe_compressed_matrix(shaped)
    if compressed is not None:
        compressed.indices.setflags(write=True)
        compressed.indptr.setflags(write=True)
        compressed.data.setflags(write=True)
        return

    if isinstance(shaped, (pd.DataFrame, pd.Series, pd.core.indexes.base.Index)):
        shaped = shaped.values

    if isinstance(shaped, pd.core.arrays.categorical.Categorical):
        shaped = shaped.codes

    if isinstance(shaped, np.ndarray):
        shaped.setflags(write=True)
        return

    raise NotImplementedError(f"unfreeze of {shaped.__class__}")


@contextmanager
def unfrozen(proper: ProperShaped) -> Iterator[None]:
    """
    Execute some in-place modification, temporarily unfreezing the ``proper`` shaped data.
    """
    is_frozen = frozen(proper)
    if is_frozen:
        unfreeze(proper)

    try:
        yield

    finally:
        if is_frozen:
            freeze(proper)


@utd.expand_doc()
def to_numpy_matrix(
    matrix: Matrix,
    *,
    default_layout: str = "row_major",
    copy: bool = False,
    only_extract: bool = False,
) -> NumpyMatrix:
    """
    Convert any :py:const:`Matrix` to a dense 2-dimensional :py:const:`NumpyMatrix`.

    If ``copy`` (default: {copy}), a copy of the data is returned even if no conversion needed to be
    done.

    If ``only_extract`` (default: {only_extract}), then assert this only extracts the data inside
    some pandas data.

    If the data is in some strange sparse format, use ``default_layout`` (default: {default_layout})
    to decide whether to return it in ``row_major`` (CSR) or ``column_major`` (CSC) layout.
    """
    assert default_layout in ("row_major", "column_major")

    sparse = maybe_sparse_matrix(matrix)
    if sparse is not None:
        assert not only_extract
        with utm.timed_step("sparse.toarray"):
            utm.timed_parameters(results=sparse.shape[0], elements=sparse.shape[1])
            layout = matrix_layout(sparse) or default_layout
            if layout == "row_major":
                order = "C"
            else:
                order = "F"
            dense = sparse.toarray(order=order)
    else:
        dense = mustbe_numpy_matrix(to_proper_matrix(matrix))

    if copy and id(dense) == id(matrix):
        dense = np.copy(dense)

    return mustbe_numpy_matrix(dense)


@utd.expand_doc()
def to_numpy_vector(
    shaped: Shaped,
    *,
    copy: bool = False,
    only_extract: bool = False,
) -> NumpyVector:
    """
    Convert any :py:const:`Vector`, or a :py:const:`Matrix` where one of the dimensions has size
    one, to a :py:const:`NumpyVector`.

    If ``copy`` (default: {copy}), a copy of the data is returned even if no conversion needed to be
    done.

    If ``only_extract`` (default: {only_extract}), then assert this only extracts the data inside
    some pandas data.
    """
    if not hasattr(shaped, "ndim"):
        dense = np.array(shaped)

    elif is_1d(shaped):
        if isinstance(shaped, (pd.DataFrame, pd.Series, pd.core.indexes.base.Index)):
            shaped = shaped.values

        if isinstance(shaped, pd.core.arrays.categorical.Categorical):
            shaped = np.array(shaped)

        dense = shaped

    else:
        assert is_2d(shaped)
        assert shaped.shape[0] == 1 or shaped.shape[1] == 1  # type: ignore
        dense = to_numpy_matrix(shaped, copy=copy, only_extract=only_extract)  # type: ignore
        dense = np.reshape(dense, -1)

    if copy and id(dense) == id(shaped):
        dense = np.copy(dense)

    return mustbe_numpy_vector(dense)


#: Which flag indicates efficient 2D dense matrix layout.
DENSE_FAST_FLAG = {"column_major": "F_CONTIGUOUS", "row_major": "C_CONTIGUOUS"}

#: Which format indicates efficient 2D sparse matrix layout.
SPARSE_FAST_FORMAT = {"column_major": "csc", "row_major": "csr"}

#: Which format indicates inefficient 2D sparse matrix layout.
SPARSE_SLOW_FORMAT = {"column_major": "csr", "row_major": "csc"}

#: The layout by the ``axis`` parameter.
LAYOUT_OF_AXIS = ("row_major", "column_major")

#: When reducing data, get results ``per`` row or column (by the ``axis`` parameter).
PER_OF_AXIS = ("row", "column")


def is_layout(matrix: Matrix, layout: Optional[str]) -> bool:
    """
    Test whether the ``matrix`` is arranged according to the ``layout``.

    This will always succeed if the ``layout`` is ``None``.
    """
    if layout is None:
        return True

    if matrix.shape[0] == 1 or matrix.shape[1] == 1:
        return True

    sparse = maybe_sparse_matrix(matrix)
    if sparse is not None:
        return sparse.getformat() == SPARSE_FAST_FORMAT[layout]

    dense = to_numpy_matrix(matrix, only_extract=True)
    return dense.flags[DENSE_FAST_FLAG[layout]]  # type: ignore


def shaped_dtype(shaped: Shaped) -> str:
    """
    Return the data type of the element of shaped data.
    """
    frame = maybe_pandas_frame(shaped)
    if frame is not None:
        shaped = frame.values

    compressed = maybe_compressed_matrix(shaped)
    if compressed is not None:
        shaped = compressed.data

    series = maybe_pandas_series(shaped)
    if series is not None:
        shaped = series.values

    if isinstance(shaped, pd.core.arrays.categorical.Categorical):
        shaped = np.array(shaped)

    if isinstance(shaped, np.ndarray):
        return str(shaped.dtype)

    raise AssertionError(f"unexpected shaped type: {shaped.__class__.__qualname__}")


def matrix_layout(matrix: Matrix) -> Optional[str]:
    """
    Return which layout the ``matrix`` is arranged by (``row_major`` or ``column_major``).

    If the data is in some strange sparse format, returns ``None``.
    """
    sparse = maybe_sparse_matrix(matrix)
    if sparse is not None:
        for layout, sparse_format in SPARSE_FAST_FORMAT.items():
            if sparse.getformat() == sparse_format:
                return layout
        return None

    dense = to_numpy_matrix(matrix, only_extract=True)
    for layout, flag in DENSE_FAST_FLAG.items():
        if dense.flags[flag]:  # type: ignore
            return layout

    return None


def is_contiguous(vector: Vector) -> bool:
    """
    Return whether the ``vector`` is contiguous in memory.

    This is only ``True`` for a dense vector.
    """
    if isinstance(vector, (pd.DataFrame, pd.Series, pd.core.indexes.base.Index)):
        vector = vector.values

    if isinstance(vector, pd.core.arrays.categorical.Categorical):
        vector = np.array(vector)

    dense = maybe_numpy_vector(vector)
    if dense is None:
        return False

    return dense.flags.c_contiguous and dense.flags.f_contiguous


def to_contiguous(vector: Vector, *, copy: bool = False) -> NumpyVector:
    """
    Return the ``vector`` in contiguous (dense) format.

    If ``copy`` (default: {copy}), a copy of the data is returned even if no conversion needed to be
    done.
    """
    dense = to_numpy_vector(vector, copy=copy)
    if not dense.flags.c_contiguous or not dense.flags.f_contiguous:
        dense = np.copy(dense)
        assert dense.flags.c_contiguous and dense.flags.f_contiguous
    return dense


def mustbe_canonical(shaped: Shaped) -> None:
    """
    Assert that some data is in canonical format.

    For numpy matrices or vectors, this means the data is contiguous (for matrices, in either
    row-major or column-major order).

    For sparse matrices, it means the data is in COO format, or compressed (CSC or CSR format), with
    sorted indices and no duplicates.

    In general, we'd like all the data stored in ``AnnData`` to be canonical.
    """
    assert hasattr(shaped, "ndim"), "non-canonical shaped data: has no ndim attribute"

    if is_1d(shaped):
        is_contiguous_shape = is_contiguous(shaped)  # type: ignore
        assert is_contiguous_shape, "non-canonical vector: non-contiguous 1D data"
        return

    assert is_2d(shaped)
    matrix: Matrix = shaped  # type: ignore

    sparse = maybe_sparse_matrix(matrix)
    if sparse is None:
        assert (
            matrix_layout(matrix) is not None
        ), "non-canonical dense matrix: is not in row-major or column-major layout"
        return

    assert sparse.getformat() in ("coo", "csr", "csc"), "non-canonical sparse matrix: is not in COO, CSR or CSC format"

    if hasattr(sparse, "has_canonical_format"):
        assert getattr(sparse, "has_canonical_format"), "non-canonical sparse matrix: might have duplicate indices"

    if hasattr(sparse, "has_sorted_indices"):
        assert getattr(sparse, "has_sorted_indices"), "non-canonical sparse matrix: might have unsorted indices"

    compressed = maybe_compressed_matrix(matrix)
    if compressed is not None:
        assert is_contiguous(compressed.indptr), "non-canonical CSC/CSR matrix: indptr array is not contiguous"
        assert is_contiguous(compressed.data), "non-canonical CSC/CSR matrix: data array is not contiguous"
        assert is_contiguous(compressed.indices), "non-canonical CSC/CSR matrix: indices array is not contiguous"


def is_canonical(shaped: Shaped) -> bool:  # pylint: disable=too-many-return-statements
    """
    Return whether the data is in canonical format.

    For numpy matrices or vectors, this means the data is contiguous (for matrices, in either
    row-major or column-major order).

    For sparse matrices, it means the data is in COO format, or compressed (CSC or CSR format), with
    sorted indices and no duplicates.

    In general, we'd like all the data stored in ``AnnData`` to be canonical.
    """
    if not hasattr(shaped, "ndim"):
        return False

    if is_1d(shaped):
        return is_contiguous(shaped)  # type: ignore

    assert is_2d(shaped)
    matrix: Matrix = shaped  # type: ignore

    sparse = maybe_sparse_matrix(matrix)
    if sparse is None:
        return matrix_layout(matrix) is not None

    if sparse.getformat() not in ("coo", "csr", "csc"):
        return False

    if hasattr(sparse, "has_canonical_format") and not getattr(sparse, "has_canonical_format"):
        return False

    if hasattr(sparse, "has_sorted_indices") and not getattr(sparse, "has_sorted_indices"):
        return False

    compressed = maybe_compressed_matrix(matrix)
    if compressed is not None:
        return (
            compressed.indptr[-1] == compressed.data.size
            and is_canonical(compressed.indptr)
            and is_canonical(compressed.data)
            and is_canonical(compressed.indices)
        )

    return True


def eliminate_zeros(compressed: CompressedMatrix) -> None:
    """
    Eliminate zeros in a compressed matrix.
    """
    with utm.timed_step("sparse.eliminate_zeros"):
        with unfrozen(compressed):
            utm.timed_parameters(before=compressed.nnz)
            compressed.eliminate_zeros()
            utm.timed_parameters(after=compressed.nnz)


def sort_indices(compressed: CompressedMatrix) -> None:
    """
    Ensure the indices are sorted in each row/column.
    """
    with utm.timed_step("sparse.sort_indices"):
        with unfrozen(compressed):
            utm.timed_parameters(before=compressed.nnz)
            compressed.sort_indices()
            utm.timed_parameters(after=compressed.nnz)


def sum_duplicates(compressed: CompressedMatrix) -> None:
    """
    Eliminate duplicates in a compressed matrix.
    """
    with utm.timed_step("sparse.sum_duplicates"):
        with unfrozen(compressed):
            utm.timed_parameters(before=compressed.nnz)
            compressed.sum_duplicates()
            utm.timed_parameters(after=compressed.nnz)


def shaped_checksum(shaped: Shaped) -> float:
    """
    Return a checksum of the contents of ``shaped`` data (for debugging reproducibility).
    """
    if is_1d(shaped):
        values = to_numpy_vector(shaped)
    else:
        values = to_numpy_matrix(shaped).ravel()  # type: ignore
    return np.nansum(values.astype("float64") * (1 + np.arange(len(values))))



"""
Downsample
----------
"""

from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "downsample_cells",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def downsample_cells(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    downsample_min_cell_quantile: float = pr.downsample_min_cell_quantile,
    downsample_min_samples: float = pr.downsample_min_samples,
    downsample_max_cell_quantile: float = pr.downsample_max_cell_quantile,
    inplace: bool = True,
    random_seed: int,
) -> Optional[Tuple[int, ut.PandasFrame]]:
    """
    Downsample the values of ``what`` (default: {what}) data.

    Downsampling is an effective way to get the same number of samples in multiple cells
    (that is, the same number of total UMIs in multiple cells), and serves as an alternative to
    normalization (e.g., working with UMI fractions instead of raw UMI counts).

    Downsampling is especially important when computing correlations between cells. When there is
    high variance between the total UMI count in different cells, then normalization will return
    higher correlation values between cells with a higher total UMI count, which will result in an
    inflated estimation of their similarity to other cells. Downsampling avoids this effect.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    **Returns**

    Unstructured Annotations
        ``downsample_samples``
            The target total number of samples in each downsampled cell.

    Variable-Observation (Gene-Cell) Annotations
        ``downsampled``
            The downsampled data where the total number of samples in each cell is at most
            ``downsample_samples``.

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns ``None``. Otherwise this
    is returned as a tuple with the samples and a pandas data frame (indexed by the cell and gene names).

    **Computation Parameters**

    1. Compute the total samples in each cell.

    2. Decide on the value to downsample to. We would like all cells to end up with at least some
       reasonable number of samples (total UMIs) ``downsample_min_samples`` (default:
       {downsample_min_samples}). We'd also like all (most) cells to end up with the highest
       reasonable downsampled total number of samples, so if possible we increase the number of
       samples, as long as at most ``downsample_min_cell_quantile`` (default:
       {downsample_min_cell_quantile}) cells will have lower number of samples. We'd also like all
       (most) cells to end up with the same downsampled total number of samples, so if we have to we
       decrease the number of samples to ensure at most ``downsample_max_cell_quantile`` (default:
       {downsample_max_cell_quantile}) cells will have a lower number of samples.

    3. Downsample each cell so that it has at most the selected number of samples. Specify a non-zero
       ``random_seed`` to make this reproducible.
    """
    total_per_cell = ut.get_o_numpy(adata, what, sum=True)

    samples = int(
        round(
            min(
                max(downsample_min_samples, np.quantile(total_per_cell, downsample_min_cell_quantile)),
                np.quantile(total_per_cell, downsample_max_cell_quantile),
            )
        )
    )

    ut.log_calc("samples", samples)

    data = ut.get_vo_proper(adata, what, layout="row_major")
    assert ut.shaped_dtype(data) == "float32"
    downsampled = ut.downsample_matrix(data, per="row", samples=samples, random_seed=random_seed)
    if inplace:
        ut.set_vo_data(adata, "downsampled", downsampled)
        ut.set_m_data(adata, "downsample_samples", samples)
        return None

    return samples, ut.to_pandas_frame(downsampled, index=adata.obs_names, columns=adata.var_names)



"""
Mark (Genes)
------------

Even when working with "clean" data, computing metacells requires marking genes for special treatment, specifically
"lateral" genes which should not be selected for computing metacells and "noisy" genes which should be neither selected
as nor be used to detect deviant (outlier) cells in the metacells.

Proper choice of lateral and noisy genes is mandatory for high-quality metacells generation. However, this choice can't
be fully automated, and therefore relies on the analyst to manually provide the list of genes.
"""

from re import Pattern
from typing import Collection
from typing import Dict
from typing import Optional
from typing import Union

from anndata import AnnData  # type: ignore

import metacells.tools as tl
import metacells.utilities as ut

__all__ = [
    "mark_lateral_genes",
    "mark_noisy_genes",
    "mark_select_genes",
    "mark_ignored_genes",
    "mark_essential_genes",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def mark_lateral_genes(
    adata: AnnData,
    *,
    lateral_gene_names: Optional[Collection[str]] = None,
    lateral_gene_patterns: Optional[Collection[Union[str, Pattern]]] = None,
    op: str = "set",
) -> None:
    """
    Mark a subset of the genes as "lateral", that is, prevent them from being selected for computing metacells.

    Lateral genes are still used to detect outliers, that is, the cells in resulting metacells should still have
    similar expression level for such genes.

    Depending on ``op``, this will either ``set`` (override/create) the mask, ``add`` (to an existing mask), or
    ``remove`` (from an existing mask).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    Sets the following in the data:

    Variable (gene) annotations:
        ``lateral_gene``
            A mask of the "lateral" genes.

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.named.find_named_genes` to also mark lateral genes based on their name, using
       the ``lateral_gene_names`` (default: {lateral_gene_names}), ``lateral_gene_patterns`` (default:
       {lateral_gene_patterns}) and ``op`` (default: {op}).
    """
    tl.find_named_genes(adata, names=lateral_gene_names, patterns=lateral_gene_patterns, to="lateral_gene", op=op)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def mark_noisy_genes(
    adata: AnnData,
    *,
    noisy_gene_names: Optional[Collection[str]] = None,
    noisy_gene_patterns: Optional[Collection[Union[str, Pattern]]] = None,
    op: str = "set",
) -> None:
    """
    Mark a subset of the genes as "noisy", that is, prevent them from being used for detecting deviant (outlier) cells.

    Depending on ``op``, this will either ``set`` (override/create) the mask, ``add`` (to an existing mask), or
    ``remove`` (from an existing mask).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    Sets the following in the data:

    Variable (gene) annotations:
        ``noisy_gene``
            A mask of the "noisy" genes.

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.named.find_named_genes` to also mark noisy genes based on their name, using
       the ``noisy_gene_names`` (default: {noisy_gene_names}), ``noisy_gene_patterns`` (default:
       {noisy_gene_patterns}) and ``op`` (default: {op}).
    """
    tl.find_named_genes(adata, names=noisy_gene_names, patterns=noisy_gene_patterns, to="noisy_gene", op=op)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def mark_select_genes(
    adata: AnnData,
    *,
    select_gene_names: Optional[Collection[str]] = None,
    select_gene_patterns: Optional[Collection[Union[str, Pattern]]] = None,
    op: str = "set",
) -> None:
    """
    Mark a subset of the genes as "select", that is, force them to be used when computing metacell.

    If this is done, then it overrides the default gene selection mechanism, forcing the algorithm to use a fixed set of
    genes. In general, this will result in lower-quality metacells, especially when using the divide-and-conquer
    algorithm, so **do not use this unless you really know what you are doing**.

    Depending on ``op``, this will either ``set`` (override/create) the mask, ``add`` (to an existing mask), or
    ``remove`` (from an existing mask).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    Sets the following in the data:

    Variable (gene) annotations:
        ``select_gene``
            A mask of the "select" genes.

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.named.find_named_genes` to select genes based on their name, using
       the ``select_gene_names`` (default: {select_gene_names}), ``select_gene_patterns`` (default:
       {select_gene_patterns}) and ``op`` (default: {op}).
    """
    tl.find_named_genes(adata, names=select_gene_names, patterns=select_gene_patterns, to="select_gene", op=op)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def mark_ignored_genes(
    adata: AnnData,
    *,
    ignored_gene_names: Optional[Collection[str]] = None,
    ignored_gene_patterns: Optional[Collection[Union[str, Pattern]]] = None,
    ignored_gene_names_of_types: Optional[Dict[str, Collection[str]]] = None,
    ignored_gene_patterns_of_types: Optional[Dict[str, Collection[str]]] = None,
    op: str = "set",
) -> None:
    """
    Mark a subset of the genes as "ignored", that is, do not attempt to match them when projecting this (query)
    data onto an atlas.

    Depending on ``op``, this will either ``set`` (override/create) the mask(s), ``add`` (to an existing mask(s)), or
    ``remove`` (from an existing mask(s)).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    Sets the following in the data:

    Variable (gene) annotations:
        ``ignored_gene``
            A mask of the "ignored" genes for all the query metacells regardless of their type.

        ``ignored_gene_of_<type>``
            A mask of the "ignored" genes for query metacells that are assigned a specific type.

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.named.find_named_genes` to ignore genes based on their name, using
       the ``ignored_gene_names`` (default: {ignored_gene_names}), ``ignored_gene_patterns`` (default:
       {ignored_gene_patterns}) and ``op`` (default: {op}).

    2. Similarly for each type specified in ``ignored_gene_names_of_types`` and/or ``ignored_gene_patterns_of_types``.
    """
    tl.find_named_genes(adata, names=ignored_gene_names, patterns=ignored_gene_patterns, to="ignored_gene", op=op)

    ignored_gene_names_of_types = ignored_gene_names_of_types or {}
    ignored_gene_patterns_of_types = ignored_gene_patterns_of_types or {}
    types = set(ignored_gene_names_of_types.keys()) | set(ignored_gene_patterns_of_types.keys())

    for type_name in types:
        tl.find_named_genes(
            adata,
            names=ignored_gene_names_of_types.get(type_name),
            patterns=ignored_gene_patterns_of_types.get(type_name),
            to=f"ignored_gene_of_{type_name}",
            op=op,
        )


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def mark_essential_genes(
    adata: AnnData,
    *,
    essential_gene_names_of_types: Optional[Dict[str, Collection[str]]] = None,
    essential_gene_patterns_of_types: Optional[Dict[str, Collection[str]]] = None,
    op: str = "set",
) -> None:
    """
    Mark a subset of the genes as "essential", that is, require that (most of them) will be projected "well" to accept a
    query metacell as having the same type as the atlas projection.

    Depending on ``op``, this will either ``set`` (override/create) the mask(s), ``add`` (to an existing mask(s)), or
    ``remove`` (from an existing mask(s)).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    Sets the following in the data:

    Variable (gene) annotations:
        ``essential_gene_of_<type>``
            A mask of the "essential" genes for query metacells to be assigned a specific type.

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.named.find_named_genes` to ignore genes based on their name, using the
       ``essential_gene_names_of_types`` (default: {essential_gene_names_of_types}),
       ``essential_gene_patterns_of_types`` (default: {essential_gene_patterns_of_types}) and ``op`` (default: {op}).
    """
    essential_gene_names_of_types = essential_gene_names_of_types or {}
    essential_gene_patterns_of_types = essential_gene_patterns_of_types or {}
    types = set(essential_gene_names_of_types.keys()) | set(essential_gene_patterns_of_types.keys())

    for type_name in types:
        tl.find_named_genes(
            adata,
            names=essential_gene_names_of_types.get(type_name),
            patterns=essential_gene_patterns_of_types.get(type_name),
            to=f"essential_gene_of_{type_name}",
            op=op,
        )



"""
Check for AVX2
--------------

This is only imported on X86_64 machines if we compiled the C++ extension to use AVX2 (when creating
the pre-compiled wheels). If this is run on a non-AVX2 machine, it generates a human-readable error
instead of generating an opaque segmentation fault.
"""

import os
from warnings import warn

HAS_AVX2 = None
try:
    with open("/proc/cpuinfo", encoding="utf8") if os.path.exists("/proc/cpuinfo") else os.popen("sysctl -a") as file:
        for line in file.readlines():
            if line.startswith("flags"):
                features = line.split(" ")
                HAS_AVX2 = "avx2" in features and "fma" in features
                break
except BaseException:  # pylint: disable=broad-except
    pass

if HAS_AVX2 is None:
    AVX2_MAYBE_NOT_SUPPORTED = (
        "The metacells precompiled wheel is using AVX2 FMA instructions.\n"
        "However, AVX2 FMA might not be available on this computer's processors.\n"
        "Therefore, using the metacells package might cause a segmentation violation.\n"
        "You can avoid the wheel using: pip install metacells --install-option=--native\n"
        "This will compile the metacells package on and for this computer's processors."
    )
    warn(AVX2_MAYBE_NOT_SUPPORTED)
    HAS_AVX2 = True

if not HAS_AVX2:
    AVX2_NOT_SUPPORTED = (
        "The metacells precompiled wheel is using AVX2 FMA instructions.\n"
        "However, AVX2 FMA is not available on this computer's processors.\n"
        "You can avoid the wheel using: pip install metacells --install-option=--native\n"
        "This will compile the metacells package on and for this computer's processors."
    )
    raise ImportError(AVX2_NOT_SUPPORTED)



"""
Noisy Lonely
------------
"""

from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut
from metacells.tools.downsample import downsample_cells
from metacells.tools.filter import filter_data
from metacells.tools.high import find_high_normalized_variance_genes
from metacells.tools.high import find_high_total_genes
from metacells.tools.similarity import compute_var_var_similarity

__all__ = [
    "find_bursty_lonely_genes",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def find_bursty_lonely_genes(  # pylint: disable=too-many-statements
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    max_sampled_cells: int = pr.bursty_lonely_max_sampled_cells,
    downsample_min_samples: int = pr.bursty_lonely_downsample_min_samples,
    downsample_min_cell_quantile: float = pr.bursty_lonely_downsample_max_cell_quantile,
    downsample_max_cell_quantile: float = pr.bursty_lonely_downsample_min_cell_quantile,
    min_gene_total: int = pr.bursty_lonely_min_gene_total,
    min_gene_normalized_variance: float = pr.bursty_lonely_min_gene_normalized_variance,
    max_gene_similarity: float = pr.bursty_lonely_max_gene_similarity,
    inplace: bool = True,
    random_seed: int,
) -> Optional[ut.PandasSeries]:
    """
    Detect "bursty lonely" genes based on ``what`` (default: {what}) data.

    Return the indices of genes which are "bursty" (have high variance compared to their mean) and
    also "lonely" (have low correlation with all other genes). Such genes should be excluded since
    they will never meaningfully help us compute groups, and will actively cause profiles to be
    considered "deviants".

    Noisy genes have high expression and variance. Lonely genes have no (or low) correlations with
    any other gene. Noisy lonely genes tend to throw off clustering algorithms. In general, such
    algorithms try to group together cells with the same overall biological state. Since the genes
    are lonely, they don't contribute towards this goal. Since they are bursty, they actively hamper
    this, because they make cells which are otherwise similar appear different (just for this lonely
    gene).

    It is therefore useful to explicitly identify, in a pre-processing step, the (few) such genes,
    and exclude them from the rest of the analysis.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    **Returns**

    Variable (Gene) Annotations
        ``bursty_lonely_genes``
            A boolean mask indicating whether each gene was found to be a "bursty lonely" gene.

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns
    ``None``. Otherwise this is returned as a pandas series (indexed by the variable names).

    **Computation Parameters**

    1. If we have more than ``max_sampled_cells`` (default: {max_sampled_cells}), pick this number
       of random cells. Specify a non-zero random seed to make this reproducible.

    2. Invoke :py:func:`metacells.tools.downsample.downsample_cells` to downsample the cells to the
       same total number of UMIs, using the ``downsample_min_samples`` (default:
       {downsample_min_samples}), ``downsample_min_cell_quantile`` (default:
       {downsample_min_cell_quantile}), ``downsample_max_cell_quantile`` (default:
       {downsample_max_cell_quantile}) and the ``random_seed``.

    3. Find "bursty" genes which have a total number of UMIs of at least ``min_gene_total`` (default:
       {min_gene_total}) and a normalized variance of at least ``min_gene_normalized_variance``
       (default: ``min_gene_normalized_variance``).

    4. Cross-correlate the bursty genes.

    5. Find the bursty "lonely" genes whose maximal correlation is at most
       ``max_gene_similarity`` (default: {max_gene_similarity}) with all other genes.
    """
    if max_sampled_cells < adata.n_obs:
        np.random.seed(random_seed)
        cell_indices = np.random.choice(np.arange(adata.n_obs), size=max_sampled_cells, replace=False)
        sdata = ut.slice(adata, obs=cell_indices, name=".sampled", top_level=False)
    else:
        sdata = ut.copy_adata(adata, top_level=False)

    track_var: Optional[str] = "sampled_gene_index"

    downsample_cells(
        sdata,
        what,
        downsample_min_samples=downsample_min_samples,
        downsample_min_cell_quantile=downsample_min_cell_quantile,
        downsample_max_cell_quantile=downsample_max_cell_quantile,
        random_seed=random_seed,
    )

    find_high_total_genes(sdata, "downsampled", min_gene_total=min_gene_total)

    results = filter_data(
        sdata, name="high_total", top_level=False, track_var=track_var, var_masks=["&high_total_gene"]
    )
    track_var = None
    assert results is not None
    ht_data = results[0]

    bursty_lonely_genes_mask = np.full(adata.n_vars, False)

    if ht_data is not None:
        ht_genes_count = ht_data.shape[1]

        ht_gene_ht_gene_similarity_frame = compute_var_var_similarity(
            ht_data, "downsampled", inplace=False, reproducible=(random_seed != 0)
        )
        assert ht_gene_ht_gene_similarity_frame is not None

        ht_gene_ht_gene_similarity_matrix = ut.to_numpy_matrix(ht_gene_ht_gene_similarity_frame, only_extract=True)
        ht_gene_ht_gene_similarity_matrix = ut.to_layout(
            ht_gene_ht_gene_similarity_matrix, layout="row_major", symmetric=True
        )
        np.fill_diagonal(ht_gene_ht_gene_similarity_matrix, -1)

        htv_mask_series = find_high_normalized_variance_genes(
            ht_data, "downsampled", min_gene_normalized_variance=min_gene_normalized_variance, inplace=False
        )
        assert htv_mask_series is not None
        htv_mask = ut.to_numpy_vector(htv_mask_series)

        htv_genes_count = np.sum(htv_mask)
        assert htv_genes_count <= ht_genes_count

        if htv_genes_count > 0:
            htv_gene_ht_gene_similarity_matrix = ht_gene_ht_gene_similarity_matrix[htv_mask, :]
            assert ut.is_layout(htv_gene_ht_gene_similarity_matrix, "row_major")
            assert htv_gene_ht_gene_similarity_matrix.shape == (htv_genes_count, ht_genes_count)

            max_similarity_of_htv_genes = ut.max_per(htv_gene_ht_gene_similarity_matrix, per="row")
            htvl_mask = max_similarity_of_htv_genes <= max_gene_similarity
            htvl_genes_count = np.sum(htvl_mask)
            ut.log_calc("bursty_lonely_genes_count", htvl_genes_count)

            if htvl_genes_count > 0:
                base_index_of_ht_genes = ut.get_v_numpy(ht_data, "sampled_gene_index")
                base_index_of_htv_genes = base_index_of_ht_genes[htv_mask]
                base_index_of_htvl_genes = base_index_of_htv_genes[htvl_mask]

                bursty_lonely_genes_mask[base_index_of_htvl_genes] = True

                htvl_gene_ht_gene_similarity_matrix = htv_gene_ht_gene_similarity_matrix[htvl_mask, :]
                htvl_gene_ht_gene_similarity_matrix = ut.to_layout(
                    htvl_gene_ht_gene_similarity_matrix, layout="row_major"
                )
                assert htvl_gene_ht_gene_similarity_matrix.shape == (htvl_genes_count, ht_genes_count)

                if ut.logging_calc():
                    i_gene_totals = ut.get_v_numpy(sdata, "downsampled", sum=True)
                    ht_mask = ut.get_v_numpy(sdata, "high_total_gene")
                    i_total = np.sum(i_gene_totals)
                    htvl_gene_totals = i_gene_totals[ht_mask][htv_mask][htvl_mask]
                    top_similarity_of_htvl_genes = ut.top_per(htvl_gene_ht_gene_similarity_matrix, 10, per="row")
                    for htvl_index, gene_index in enumerate(base_index_of_htvl_genes):
                        gene_name = adata.var_names[gene_index]
                        gene_total = htvl_gene_totals[htvl_index]
                        gene_percent = 100 * gene_total / i_total
                        similar_ht_values = ut.to_numpy_vector(top_similarity_of_htvl_genes[htvl_index, :])  #
                        assert len(similar_ht_values) == ht_genes_count
                        top_similar_ht_mask = similar_ht_values > 0
                        top_similar_ht_values = similar_ht_values[top_similar_ht_mask]
                        top_similar_ht_indices = base_index_of_ht_genes[top_similar_ht_mask]
                        top_similar_ht_names = adata.var_names[top_similar_ht_indices]
                        ut.log_calc(
                            f"- {gene_name}",
                            f"total downsampled UMIs: {gene_total} "
                            + f"({gene_percent:.4g}%), correlated with: "
                            + ", ".join(
                                [
                                    f"{similar_gene_name}: {similar_gene_value:.4g}"
                                    for similar_gene_value, similar_gene_name in reversed(
                                        sorted(zip(top_similar_ht_values, top_similar_ht_names))
                                    )
                                ]
                            ),
                        )

    if ut.logging_calc():
        ut.log_calc("bursty_lonely_gene_names", sorted(list(adata.var_names[bursty_lonely_genes_mask])))

    if inplace:
        ut.set_v_data(adata, "bursty_lonely_gene", bursty_lonely_genes_mask)
        return None

    ut.log_return("bursty_lonely_genes", bursty_lonely_genes_mask)
    return ut.to_pandas_series(bursty_lonely_genes_mask, index=adata.var_names)



"""
Logging
-------

This provides a useful formatter which includes high-resolution time and thread names, and a set of
utility functions for effective logging of operations on annotation data.

Collection of log messages is mostly automated by wrapping relevant function calls and tracing the
setting and getting of data via the :py:mod:`metacells.utilities.annotation` accessors, with the
occasional explicit logging of a notable intermediate calculated value via :py:func:`log_calc`.

The ging is picking the correct level for each log message. This module provides the following log
levels which hopefully provide the end user with a reasonable amount of control:

* ``INFO`` will log only setting of the final results as annotations within the top-level
  ``AnnData`` object(s).

* ``STEP`` will also log the top-level algorithm step(s), which give a very basic insight into
  what was executed.

* ``PARAM`` will also log the parameters of these steps, which may be important when tuning the
  behavior of the system for different data sets.

* ``CALC`` will also log notable intermediate calculated results, which again may be important
  when tuning the behavior of the system for different data sets.

* ``DEBUG`` pulls all the stops and logs all the above, not only for the top-level steps, but
  also for the nested processing steps. This results in a rather large log file (especially
  for the recursive divide-and-conquer algorithm). You don't need this except for when
  you really need this.

To achieve this, we track for each ``AnnData`` whether it is a top-level (user visible) or a
temporary data object, and whether we are inside a top-level (user invoked) or a nested operation.
Accessing top-level data and invoking top-level operations is logged at the coarse logging levels,
anything else is logged at the ``DEBUG`` level.

To improve the log messages, we allow each ``AnnData`` object to have an optional name for logging
(see :py:func:`metacells.utilities.annotation.set_name` and
:py:func:`metacells.utilities.annotation.get_name`). Whenever a temporary ``AnnData`` data is
created, its name is extended by some descriptive suffix, so we get names like
``full.clean.select`` to describe the data selected from the clean data extracted out of
the full data.
"""

import importlib
import logging
import sys
import warnings
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from inspect import Parameter
from inspect import signature
from logging import DEBUG
from logging import INFO
from logging import Formatter
from logging import Logger
from logging import LogRecord
from logging import StreamHandler
from logging import getLogger
from logging import setLoggerClass
from multiprocessing import Lock
from threading import current_thread
from typing import IO
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

import numpy as np
import pandas as pd  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.utilities.documentation as utd

utp: Any = importlib.import_module("metacells.utilities.parallel")
utt: Any = importlib.import_module("metacells.utilities.typing")

__all__ = [
    "setup_logger",
    "logger",
    "CALC",
    "STEP",
    "PARAM",
    "logged",
    "top_level",
    "log_return",
    "logging_calc",
    "log_calc",
    "log_step",
    "incremental",
    "done_incrementals",
    "cancel_incrementals",
    "log_set",
    "log_get",
    "sizes_description",
    "fractions_description",
    "groups_description",
    "mask_description",
    "ratio_description",
    "progress_description",
    "fraction_description",
    "fold_description",
]


warnings.filterwarnings(
    "ignore", category=FutureWarning, message=r".*Reordering categories will always return a new Categorical object.*"
)
warnings.filterwarnings(
    "ignore",
    category=FutureWarning,
    message=r".*is_categorical is deprecated and will be removed in a future version.*",
)


class SynchronizedLogger(Logger):
    """
    A logger that synchronizes between the sub-processes.

    This ensures logging does not get garbled when when using multi-processing
    (e.g., using :py:func:`metacells.utilities.parallel.parallel_map`).
    """

    LOCK = Lock()

    def _log(self, *args: Any, **kwargs: Any) -> Any:
        with SynchronizedLogger.LOCK:
            super()._log(*args, **kwargs)


class LoggingFormatter(Formatter):
    """
    A formatter that uses a decimal point for milliseconds.
    """

    def formatTime(self, record: Any, datefmt: Optional[str] = None) -> str:
        """
        Format the time.
        """
        record_datetime = datetime.fromtimestamp(record.created)
        if datefmt is not None:
            assert False
            return record_datetime.strftime(datefmt)

        seconds = record_datetime.strftime("%Y-%m-%d %H:%M:%S")
        msecs = round(record.msecs)
        return f"{seconds}.{msecs:03d}"


#: The log level for tracing processing steps.
STEP = (1 * DEBUG + 3 * INFO) // 4

#: The log level for tracing parameters.
PARAM = (2 * DEBUG + 2 * INFO) // 4

#: The log level for tracing intermediate calculations.
CALC = (3 * DEBUG + 1 * INFO) // 4

logging.addLevelName(STEP, "STEP")
logging.addLevelName(PARAM, "PARAM")
logging.addLevelName(CALC, "CALC")


class ShortLoggingFormatter(LoggingFormatter):
    """
    Provide short level names.
    """

    #: Map the long level names to the fixed-width short level names.
    SHORT_LEVEL_NAMES = {
        "CRITICAL": "CRT",
        "CRT": "CRT",
        "ERROR": "ERR",
        "ERR": "ERR",
        "WARNING": "WRN",
        "WRN": "WRN",
        "INFO": "INF",
        "INF": "INF",
        "STEP": "STP",
        "STP": "STP",
        "PARAM": "PRM",
        "PRM": "PRM",
        "HINT": "PRM",
        "HNT": "PRM",
        "CALC": "CLC",
        "CLC": "CLC",
        "DEBUG": "DBG",
        "DBG": "DBG",
        "NOTSET": "NOT",
        "NOT": "NOT",
    }

    def format(self, record: LogRecord) -> Any:
        record.levelname = self.SHORT_LEVEL_NAMES[record.levelname]
        return LoggingFormatter.format(self, record)


# Global logger object.
LOG: Optional[Logger] = None


@utd.expand_doc()
def setup_logger(
    *,
    level: int = logging.INFO,
    to: IO = sys.stderr,
    time: bool = False,
    process: Optional[bool] = None,
    name: Optional[str] = None,
    long_level_names: Optional[bool] = None,
) -> Logger:
    """
    Setup the global :py:func:`logger`.

    .. note::

        A second call will fail as the logger will already be set up.

    If ``level`` is not specified, only ``INFO`` messages (setting values in the annotated data)
    will be logged.

    If ``to`` is not specified, the output is sent to ``sys.stderr``.

    If ``time`` (default: {time}), include a millisecond-resolution timestamp in each message.

    If ``name`` (default: {name}) is specified, it is added to each message.

    If ``process`` (default: {process}), include the (sub-)process index in each message. The name
    of the main process (thread) is replaced to ``#0`` to make it more compatible with the
    sub-process names (``#<map-index>.<sub-process-index>``).

    If ``process`` is ``None``, and if the logging level is higher than ``INFO``, and
    :py:func:`metacells.utilities.parallel.get_processors_count` is greater than one, then
    ``process`` is set - that is, it will be set if we expect to see log messages from multiple
    sub-processes.

    Logging from multiple sub-processes (e.g., using (e.g., using
    :py:func:`metacells.utilities.parallel.parallel_map`) will synchronize using a global lock so
    messages will not get garbled.

    If ``long_level_names`` (default: {long_level_names}), includes the log level in each message.
    If is ``False``, the log level names are shortened to three characters, for consistent
    formatting of indented (nested) log messages. If it is ``None``, no level names are logged at
    all.
    """
    assert utp.is_main_process()

    global LOG
    assert LOG is None

    if long_level_names is not None:
        log_format = "%(levelname)s - %(message)s"
    else:
        log_format = "%(message)s"

    if process is None:
        process = level < logging.INFO and utp.get_processors_count() > 1

    if process:
        log_format = "%(threadName)s - " + log_format
        current_thread().name = "#0"

    if name is not None:
        log_format = name + " - " + log_format
    if time:
        log_format = "%(asctime)s - " + log_format

    handler = StreamHandler(to)
    if long_level_names is False:
        handler.setFormatter(ShortLoggingFormatter(log_format))
    else:
        handler.setFormatter(LoggingFormatter(log_format))
    setLoggerClass(SynchronizedLogger)
    LOG = getLogger("metacells")
    LOG.addHandler(handler)
    LOG.setLevel(level)

    LOG.debug("PROCESSORS: %s", utp.get_processors_count())
    return LOG


def logger() -> Logger:
    """
    Access the global logger.

    If :py:func:`setup_logger` has not been called yet, this will call it using the default flags.
    You should therefore call :py:func:`setup_logger` as early as possible to ensure you don't end
    up with a misconfigured logger.
    """
    global LOG
    if LOG is None:
        LOG = setup_logger()
    return LOG


CALLABLE = TypeVar("CALLABLE")

CALL_LEVEL = 0
INDENT_LEVEL = 0
INDENT_SPACES = "  " * 1000
IS_TOP_LEVEL = True


def logged(**kwargs: Callable[[Any], Any]) -> Callable[[CALLABLE], CALLABLE]:
    """
    Automatically wrap each invocation of the decorated function with logging it. Top-level calls
    are logged using the :py:const:`STEP` log level, with parameters logged at the :py:const:`PARAM`
    log level. Nested calls are logged at the ``DEBUG`` log level.

    By default parameters are logged by simply converting them to a string, with special cases for
    ``AnnData``, callable functions, boolean masks, vectors and matrices. You can override this by
    specifying ``parameter_name=convert_value_to_logged_value`` for the specific parameter.

    Expected usage is:

    .. code:: python

        @ut.logged()
        def some_function(...):
            ...
    """
    formatter_by_name = kwargs

    def wrap(function: Callable) -> Callable:
        parameters = signature(function).parameters
        for name in formatter_by_name:
            if name not in parameters.keys():
                raise RuntimeError(
                    f"formatter specified for the unknown parameter: {name} "
                    f"for the function: {function.__module__}.{function.__qualname__}"
                )
        ordered_parameters = list(parameters.values())

        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:  # pylint: disable=too-many-branches
            global CALL_LEVEL
            global INDENT_LEVEL
            names, values = _collect_parameters(ordered_parameters, *args, **kwargs)
            adatas = _collect_adatas(values)
            new_adatas: List[AnnData] = []

            global IS_TOP_LEVEL
            old_is_top_level = IS_TOP_LEVEL

            try:
                for adata in adatas:
                    IS_TOP_LEVEL = False
                    if hasattr(adata, "__is_top_level__"):
                        IS_TOP_LEVEL = IS_TOP_LEVEL or getattr(adata, "__is_top_level__")
                    else:
                        IS_TOP_LEVEL = IS_TOP_LEVEL or CALL_LEVEL == 0
                        setattr(adata, "__is_top_level__", CALL_LEVEL == 0)
                        new_adatas.append(adata)

                if IS_TOP_LEVEL:
                    step_level = STEP
                    param_level = PARAM
                else:
                    step_level = DEBUG
                    param_level = DEBUG

                name = function.__qualname__
                if name[0] == "_":
                    name = name[1:]
                if logger().isEnabledFor(step_level):
                    logger().log(step_level, "%scall %s:", INDENT_SPACES[: 2 * INDENT_LEVEL], name)
                    INDENT_LEVEL += 1
                CALL_LEVEL += 1

                if logger().isEnabledFor(param_level):
                    for name, value in zip(names, values):
                        log_value = _format_value(value, name, formatter_by_name.get(name))
                        if log_value is not None:
                            logger().log(
                                param_level, "%swith %s: %s", INDENT_SPACES[: 2 * INDENT_LEVEL], name, log_value
                            )

                return function(*args, **kwargs)

            finally:
                if logger().isEnabledFor(step_level):
                    INDENT_LEVEL -= 1
                CALL_LEVEL -= 1
                IS_TOP_LEVEL = old_is_top_level
                for adata in new_adatas:
                    delattr(adata, "__is_top_level__")

        return wrapper

    return wrap  # type: ignore


def _collect_parameters(parameters: List[Parameter], *args: Any, **kwargs: Any) -> Tuple[List[str], List[Any]]:
    names: List[str] = []
    values: List[Any] = []

    for value, parameter in zip(args, parameters):
        names.append(parameter.name)
        values.append(value)

    for parameter in parameters[len(args) :]:
        names.append(parameter.name)
        values.append(kwargs.get(parameter.name, parameter.default))

    return names, values


def _collect_adatas(values: List[Any]) -> List[AnnData]:
    adatas: List[AnnData] = []
    for value in values:
        if isinstance(value, AnnData):
            adatas.append(value)
        if isinstance(value, list):
            adatas += _collect_adatas(value)
    return adatas


def _format_value(  # pylint: disable=too-many-return-statements,too-many-branches
    value: Any, name: str, formatter: Optional[Callable[[Any], Any]] = None
) -> Optional[str]:
    if isinstance(value, Parameter.empty.__class__):  # type: ignore
        return None

    checksum = ""
    # if utt.is_1d(value) and 'U' not in str(value.dtype) and 'o' not in str(value.dtype):
    #    checksum = ' checksum: %.20e' % utt.shaped_checksum(value)
    # elif utt.is_2d(value):
    #    checksum = ' checksum: %.20e' % utt.shaped_checksum(value)

    # import psutil
    # vm = psutil.virtual_memory()
    # used = int(round(vm.used / (1024 * 1024 * 1024)))
    # avail = int(round(vm.available / (1024 * 1024 * 1024)))
    # free = int(round(vm.free / (1024 * 1024 * 1024)))
    # avail -= free
    # checksum += f" (M: {used}U {avail}A {free}F)"

    if formatter is not None:
        value = formatter(value)
        if value is None:
            return None

    if isinstance(value, AnnData):
        aname = value.uns.get("__name__", "unnamed")

        if hasattr(value.X, "dtype"):
            dtype = getattr(value.X, "dtype")
        else:
            dtype = "unknown"

        return f"{aname} annotated data with {value.shape[0]} X {value.shape[1]} {dtype}s" + checksum

    if isinstance(value, (np.bool_, bool, str, None.__class__)):
        return str(value) + checksum

    if isinstance(value, (int, float, np.float32, np.float64, np.int32, np.int64)):
        value = float(value)
        if "random_seed" in name:
            if value == 0:
                return "0 (time)"
            return f"{value:g} (reproducible)"
        if "rank" in name:
            return f"{value:g}"
        if "fold" in name and "nnz" not in name and (-10 <= value <= 10):
            return fold_description(value)
        if "fraction" in name or "quantile" in name:
            return fraction_description(value)
        if "factor" in name and "nnz" not in name:
            return f"X {value:.4g}"
        return f"{value:g}" + checksum

    if hasattr(value, "__qualname__"):
        return getattr(value, "__qualname__") + checksum

    if (
        isinstance(value, (pd.DataFrame, pd.Series, np.ndarray))
        and value.ndim in (1, 2)
        and utt.shaped_dtype(value) == "bool"
    ):
        return mask_description(value) + checksum

    if hasattr(value, "ndim"):
        if value.ndim == 2:
            text = f"{value.__class__.__name__} {value.shape[0]} X {value.shape[1]} {utt.shaped_dtype(value)}s"
            if hasattr(value, "nnz"):
                percent = 100 * value.nnz / (value.shape[0] * value.shape[1])
                text += f" ({value.nnz} > 0, {percent:.4g}%)"
            return text + checksum

        if value.ndim == 1:
            value = utt.to_numpy_vector(value)
            if len(value) > 20:
                text = f"{len(value)} {value.dtype}s"
                return text + checksum
            value = list(value)

    if isinstance(value, list):
        if len(value) > 20 or (len(value) > 0 and hasattr(value[0], "__len__") and not isinstance(value[0], str)):
            return f"{len(value)} {value[0].__class__.__name__}s" + checksum
        texts = [
            element.uns.get("__name__", "unnamed") if isinstance(element, AnnData) else str(element)
            for element in value
        ]
        return f'[ {", ".join(texts)} ]' + checksum

    if isinstance(value, dict):
        values = list(value.values())
        if len(values) > 20 or (len(values) > 0 and hasattr(values[0], "__len__") and not isinstance(values[0], str)):
            keys = list(value.keys())
            return f"dict {_format_value(keys, 'keys')} => {_format_value(values, 'values')}"
        return str(value) + checksum

    return f"{value.__class__.__module__}.{value.__class__.__qualname__}#{id(value)}"


def top_level(adata: AnnData) -> None:
    """
    Indicate that the annotated data will be returned to the top-level caller, increasing its
    logging level.
    """
    setattr(adata, "__is_top_level__", True)


def log_return(name: str, value: Any, *, formatter: Optional[Callable[[Any], Any]] = None) -> bool:
    """
    Log a ``value`` returned from a function with some ``name``.

    If ``formatter`` is specified, use it to override the default logged value formatting.
    """
    if CALL_LEVEL == 0:
        top_log_level = INFO
    else:
        top_log_level = CALC

    return _log_value(name, value, "return", None, top_log_level, formatter)


def logging_calc() -> bool:
    """
    Whether we are actually logging the intermediate calculations.
    """
    return (IS_TOP_LEVEL and logger().isEnabledFor(CALC)) or (not IS_TOP_LEVEL and logger().isEnabledFor(DEBUG))


def log_calc(name: str, value: Any = None, *, formatter: Optional[Callable[[Any], Any]] = None) -> bool:
    """
    Log an intermediate calculated ``value`` computed from a function with some ``name``.

    If ``formatter`` is specified, use it to override the default logged value formatting.
    """
    return _log_value(name, value, "calc", None, CALC, formatter)


@contextmanager
def log_step(name: str, value: Any = None, *, formatter: Optional[Callable[[Any], Any]] = None) -> Iterator[None]:
    """
    Same as :py:func:`log_calc`, but also further indent all the log messages inside the ``with``
    statement body.
    """
    global INDENT_LEVEL

    if log_calc(name, value, formatter=formatter):
        delta = 1
    else:
        delta = 0

    INDENT_LEVEL += delta
    try:
        yield
    finally:
        INDENT_LEVEL -= delta


#: Convert ``per`` to the name of the ``AnnData`` data member holding it.
MEMBER_OF_PER = {
    "m": "uns",
    "o": "obs",
    "v": "var",
    "oo": "obsp",
    "vv": "varp",
    "oa": "obsm",
    "va": "varm",
    "vo": "layers",
}


def incremental(adata: AnnData, per: str, name: str, formatter: Optional[Callable[[Any], Any]] = None) -> None:
    """
    Declare that the named annotation will be built incrementally - set and then repeatedly modified.
    """
    assert per in ("m", "v", "o", "vv", "oo", "vo", "va", "oa")
    if not hasattr(adata, "__incremental__"):
        setattr(adata, "__incremental__", {})
    by_name: Dict[str, Tuple[str, Optional[Callable[[Any], Any]]]] = getattr(adata, "__incremental__")
    # assert name not in by_name
    by_name[name] = (per, formatter)


def done_incrementals(adata: AnnData) -> None:
    """
    Declare that all the incremental values have been fully computed.
    """
    assert hasattr(adata, "__incremental__")
    by_name: Dict[str, Tuple[str, Optional[Callable[[Any], Any]]]] = getattr(adata, "__incremental__")
    setattr(adata, "__incremental__", {})

    for name, (per, formatter) in by_name.items():
        if name == "__x__":
            value = adata.X
        else:
            annotation = getattr(adata, MEMBER_OF_PER[per])
            if name not in annotation:
                raise RuntimeError(f"missing the incremental data: {name}")
            value = annotation[name]
        log_set(adata, per, name, value, formatter=formatter)


def cancel_incrementals(adata: AnnData) -> None:
    """
    Cancel tracking incremental annotations.
    """
    assert hasattr(adata, "__incremental__")
    delattr(adata, "__incremental__")


def log_set(
    adata: AnnData, per: str, name: str, value: Any, *, formatter: Optional[Callable[[Any], Any]] = None
) -> bool:
    """
    Log setting some annotated data.
    """
    assert per in ("m", "v", "o", "vv", "oo", "vo", "va", "oa")

    is_top_level = (
        hasattr(adata, "__is_top_level__")
        and getattr(adata, "__is_top_level__")
        and (not hasattr(adata, "__incremental__") or name not in getattr(adata, "__incremental__"))
    )

    adata_name = adata.uns.get("__name__", "unnamed")
    if name == "__x__":
        name = f"{adata_name}.X"
    else:
        member_name = MEMBER_OF_PER[per]
        name = f"{adata_name}.{member_name}[{name}]"
    return _log_value(name, value, "set", is_top_level, INFO, formatter)


def log_get(
    adata: AnnData, per: str, name: Any, value: Any, *, formatter: Optional[Callable[[Any], Any]] = None
) -> bool:
    """
    Log getting some annotated data.
    """
    assert per in ("m", "v", "o", "vv", "oo", "vo", "va", "oa")

    is_top_level = (
        hasattr(adata, "__is_top_level__")
        and getattr(adata, "__is_top_level__")
        and (
            not hasattr(adata, "__incremental__")
            or (isinstance(name, str) and name not in getattr(adata, "__incremental__"))
        )
    )

    adata_name = adata.uns.get("__name__", "unnamed")
    if isinstance(name, str) and name == "__x__":
        name = f"{adata_name}.X"
    else:
        member_name = MEMBER_OF_PER[per]
        if not isinstance(name, str):
            name = "<data>"
        name = f"{adata_name}.{member_name}[{name}]"

    return _log_value(name, value, "get", is_top_level, CALC, formatter)


def _log_value(
    name: str,
    value: Any,
    kind: str,
    is_top_level: Optional[bool],
    top_log_level: int,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> bool:
    if is_top_level is None:
        is_top_level = IS_TOP_LEVEL

    if is_top_level:
        level = top_log_level
    else:
        level = DEBUG

    if not logger().isEnabledFor(level):
        return False

    if value is None:
        logger().log(level, "%s%s", INDENT_SPACES[: 2 * INDENT_LEVEL], name)
    else:
        log_value = _format_value(value, name, formatter)
        if name[0] == "-":
            logger().log(level, "%s%s: %s", INDENT_SPACES[: 2 * INDENT_LEVEL], name, log_value)
        else:
            logger().log(level, "%s%s %s: %s", INDENT_SPACES[: 2 * INDENT_LEVEL], kind, name, log_value)

    return True


def sizes_description(sizes: Union[utt.Vector, str]) -> str:
    """
    Return a string for logging an array of sizes.
    """
    if isinstance(sizes, str):
        return sizes

    sizes = utt.to_numpy_vector(sizes)
    mean = np.mean(sizes)
    return f"{len(sizes)} {sizes.dtype}s with mean {mean:.4g}"


def fractions_description(sizes: Union[utt.Vector, str]) -> str:
    """
    Return a string for logging an array of fractions (between zero and one).
    """
    if isinstance(sizes, str):
        return sizes

    sizes = utt.to_numpy_vector(sizes)
    mean = np.mean(sizes)
    percent = mean * 100
    return f"{len(sizes)} {sizes.dtype}s with mean {mean:.4g} ({percent:.4g}%)"


def groups_description(groups: Union[utt.Vector, str]) -> str:
    """
    Return a string for logging an array of group indices.

    .. note::

        This assumes that the indices are consecutive, with negative values indicating "outliers".
    """
    if isinstance(groups, str):
        return groups

    groups = utt.to_numpy_vector(groups)
    groups_count = np.max(groups) + 1
    outliers_count = np.sum(groups < 0)

    if groups_count == 0:
        assert outliers_count == len(groups)
        return f"{len(groups)} {groups.dtype} elements with all outliers (100%)"

    grouped_count = len(groups) - outliers_count
    mean = grouped_count / groups_count
    return (
        ratio_description(len(groups), f"{groups.dtype} element", outliers_count, "outliers", base=False)
        + " and "
        + ratio_description(len(groups), f"{groups.dtype} element", grouped_count, "grouped")
        + f" with {groups_count} groups with mean size {mean:.4g}"
    )


def mask_description(mask: Union[str, utt.Vector, utt.Matrix]) -> str:
    """
    Return a string for logging a boolean mask.
    """
    if isinstance(mask, str):
        return mask

    if mask.dtype == "bool":
        value = "true"
    else:
        value = "positive"

    if mask.ndim == 2:
        return f"{mask.shape[0]} X {mask.shape[1]} " + ratio_description(
            mask.size, str(mask.dtype), np.sum(mask > 0), value
        )

    return ratio_description(mask.size, str(mask.dtype), np.sum(mask > 0), value)


def ratio_description(denominator: float, element: str, numerator: float, condition: str, *, base: bool = True) -> str:
    """
    Return a string for describing a ratio (including a percent representation).
    """
    assert numerator >= 0
    assert denominator > 0

    if int(numerator) == numerator:
        numerator = int(numerator)
    if int(denominator) == denominator:
        denominator = int(denominator)

    percent = (numerator * 100) / denominator
    if base:
        return f"{numerator} {condition} ({percent:.4g}%) out of {denominator} {element}s"
    return f"{numerator} {condition} ({percent:.4g}%)"


def progress_description(amount: int, index: int, element: str) -> str:
    """
    Return a string for describing progress in a loop.
    """
    assert amount > 0
    assert 0 <= index < amount

    percent = ((index + 1) * 100) / amount
    return f"#{index} out of {amount} {element}s ({percent:.4g}%)"


def fraction_description(fraction: Optional[float]) -> str:
    """
    Return a string for describing a fraction (including a percent representation).
    """
    if fraction is None:
        return "None"
    percent = fraction * 100
    return f"{fraction:.4g} ({percent:.4g}%)"


def fold_description(fold: float) -> str:
    """
    Return a string for describing a fraction (including a percent representation).
    """
    if fold is None:
        return "None"

    if fold >= 0:
        value = 2**fold
        char = "X"
    else:
        value = 2**-fold
        char = "/"

    return f"{fold:.4g} ({char} {value:.4g})"
